/**************************************************************************
 * Copyright 1994-1999 Patrick Powell, San Diego, CA <papowell@astart.com>
 **************************************************************************/
/**** HEADER *****/
#include "patchlevel.h"
 static char *const _id =
  "Version " PATCHLEVEL "\n"
  "$Id: ifhp.c,v 1.79 2002/01/23 00:32:32 papowell Exp papowell $"
  "\n"
  "Copyright 1992-2000 Patrick Powell <papowell@astart.com>"
  "\n"
  ;

#include "ifhp.h"
#include "debug.h"

/*
 * Main program:
 * 1. does setup - checks mode
 * 2. initializes connection
 * 3. initializes printer
 * 4. checks printer for idle status
 * 5. gets pagecount
 * 6.
 *   if IF mode {
 *       determines job format
 *         - binary, PCL, PostScript, Text
 *       if PCL then sets up PCL
 *       if PostScript then sets up PostScript
 *       if Text then
 *          sets up Text
 *          if text converter,  runs converter
 *       puts job out to printer
 *   }
 *   if OF mode {
 *       reads input line;
 *       if suspend string, then suspend
 *       else pass through
 *   }
 * 7. terminates job
 * 8. checks printer for end of job status
 * 9. gets page count
 */


 char *Outbuf;	/* buffer */
 int Outmax;		/* max buffer size */
 int Outlen;		/* length to output */
 char *Inbuf;	/* buffer */
 int Inmax;		/* max buffer size */
 int Inlen;		/* total input */

 char *PJL_UEL_str =  "\033%-12345X";
 char *PJL_str =  "@PJL\n";
 char *PJL_INFO_PAGECOUNT_str = "@PJL INFO PAGECOUNT \n";
 char Jobname[SMALLBUFFER];
 char End_ready_msg[SMALLBUFFER];
 static void Pjl_displayname(void);


 int main(int argc,char *argv[], char *envp[])
{
	struct stat statb;
	char *s, *dump;
	int i, fd, c, do_pagecount, pagecount_ps, pagecount_pjl;
	struct line_list l;

	/* make sure that the NSLPATH cannot be use to compromisze root
	 * run executables
	 */

	Init_line_list(&l);

	Trace_on_stderr = 0;
	Debug = 0;
	dump = 0;

    /* FATAL("%s",1); */

	Argc = argc;
	Argv = argv;
	Envp = envp;
	if( argc ){
		Name = argv[0];
		if( (s = safestrrchr( Name, '/' )) ){
			Name = s+1;
		}
	} else {
		Name = "????";
	}

	/* paranoia,  or even tripletanoia */
    if( (getuid() == 0 || geteuid() == 0) && getenv("NLSPATH") ){
#if defined(HAVE_UNSETENV)
		unsetenv("NLSPATH");
#elif defined(HAVE_SETENV)
		setenv("NLSPATH","",1);
#elif defined(HAVE_PUTENV)
		putenv("NLSPATH=");
#else
#error need to have one of unsetenv(), setenv(), or putenv()
#endif
	}

#if HAVE_LOCALE_H
	setlocale(LC_ALL, "");
#endif

	bindtextdomain (PACKAGE, LOCALEDIR);
	textdomain (PACKAGE);

	time( &Start_time );
	/* do setup */

	/* set up the accounting FD for output */
	if( fstat(0,&statb) == -1 ){
		Errorcode = JABORT;
		FATAL( _("ifhp: stdin is not open"));
	}
	if( fstat(1,&statb) == -1 ){
		Errorcode = JABORT;
		FATAL( _("ifhp: stdout is not open"));
	}
	if( fstat(2,&statb) == -1 ){
		if( (fd = open( "/dev/null", O_WRONLY )) != 2 ){
			Errorcode = JABORT;
			LOGERR_DIE( _("ifhp: open /dev/null failed") );
		}
	}

	/* initialize input and output buffers */
	Init_outbuf();
	Init_inbuf();

	/* check the environment variables */
	if( (s = getenv("PRINTCAP_ENTRY")) ){
		DEBUG3("main: PRINTCAP_ENTRY '%s'", s );
		Split( &l, s, ":", 1, Value_sep, 1, 1, 0  );
		if(DEBUGL3)Dump_line_list("main: PRINTCAP_ENTRY", &l );
		if( (s = Find_str_value(&l,"ifhp",Value_sep)) ){
			DEBUG1("main: getting PRINTCAP info '%s'", s );
			Fix_special_user_opts("PRINTCAP ifhp Topts", &Topts, &Unsorted_Topts, s );
		}
		Printcap_lp = safestrdup( Find_str_value(&l,"lp",Value_sep), __FILE__,__LINE__ );
		DEBUG3("main: Printcap_lp '%s'", Printcap_lp );
		Free_line_list(&l);
	}

	/* get the argument lists */
	getargs( argc, argv );
	/* we extract the debug information */
	Trace_on_stderr = Find_flag_value( &Topts, "trace", Value_sep );
	if( (s = Find_str_value( &Topts, "debug", Value_sep )) ){
		i = atoi(s);
		if( i > Debug ) Debug = i;
	}
	DEBUG1("main: Debug '%d'", Debug );
	/* check for config file */
	if( !(Config_file = Find_str_value(&Topts,"config",Value_sep)) ){
		Config_file = IFHP_CONF;
	}

	/* find the model id. First we check the Toptions, then
	 * we check to see if we use a command line flag
	 */
	Model_id = Find_str_value(&Topts,"model",Value_sep);
	dump = Find_exists_value(&Topts,"dump",Value_sep);
	DEBUG1("main: dump %s",dump );
	if( !Model_id &&
		(s = Find_str_value(&Topts,"model_from_option",Value_sep)) ){
		DEBUG1("main: checking model_from_option '%s'", s );
		for( ; !Model_id && *s; ++s ){
			c = cval(s);
			if( isupper(c) ) Model_id = Upperopts[c-'A'];
			if( islower(c) ) Model_id = Loweropts[c-'a'];
		}
	}
	DEBUG1("main: Model_id '%s'", Model_id );
	DEBUG1("main: LANG '%s', 'TRANSLATE TEST' is '%s'",
		getenv("LANG"), _("TRANSLATE TEST"));

	/* read the configuration file(s) */
	if( Read_file_list( &Raw, Config_file, "\n", 0, 0, 0, 0, 1, 0, 1 ) == 0 ){
		Errorcode = JABORT;
		FATAL(_("main: no readable config file found in '%s'"), Config_file );
	}
	/* no configuration? big problems */
	if( Raw.count == 0 ){
		Errorcode = JABORT;
		FATAL(_("main: no config file information in '%s'"), Config_file );
	}
	if(DEBUGL5)Dump_line_list("main: Raw", &Raw );

	/* now we get the default values */
	DEBUG1("main: scanning Raw for default, then model '%s'", Model_id );
	Select_model_info( &Model, &Raw, "default", 0, 0, 3 );
	if(DEBUGL5)Dump_line_list("main: Model after defaults", &Model );

	/* now we get the values for the printer model, overriding defaults */
	if( Model_id && cval(Model_id) && safestrcasecmp(Model_id, "default") ){
		DEBUG1("main: scanning for model '%s'", Model_id  );
		if( !Select_model_info( &Model, &Raw, Model_id, 1, 0, 3 ) ){
			Errorcode = JABORT;
			FATAL(_("main: no model '%s' information"), Model_id );
		}
		if(DEBUGL5)Dump_line_list("main: Model after Raw", &Model );
	}

	/* now we add in the Topts, overriding defaults AND printer */
	Merge_list( &Model, &Topts, Value_sep, 1, 1 );
	if(DEBUGL5)Dump_line_list("main: Model after Topts", &Model );

	/* handy way to get a dump of the configuration for the printer */
	if( dump ){
		FPRINTF(STDOUT,"Configuration for model='%s'\n",
			Model_id?Model_id:"default");
		for( i = 0; i < Model.count; ++i ){
			if( (s = Model.list[i]) ){
				FPRINTF(STDOUT, "[%3d]  %s\n", i, s );
			}
		}
		exit(0);
	}

	if( OF_Mode && (s = Find_str_value( &Model,"of_options", Value_sep )) ){ 
		Free_line_list( &l );
		Split( &l, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: of_options", &l);
		Free_line_list( &l );
		Split( &Model, s, List_sep, 1, Value_sep, 1, 1, 0 );
	}

	/* set the named parameter variables with values */
	if(DEBUGL4) Dump_line_list( "main: Model information", &Model );
	Initialize_parms(&Model, Valuelist );
	if(DEBUGL2) Dump_parms( "main: variable values", Valuelist );

	/* check for -Zlogall allowed, and set it */
	if( (s = Find_str_value( &Model, "user_opts", Value_sep)) ){
		lowercase(s);
		Split( &User_opts, s, List_sep, 1, Value_sep, 1, 1, 0  );
		if(DEBUGL4)Dump_line_list("main: user_opts", &User_opts);
		if( Find_flag_value( &Zopts, "logall", Value_sep )
			&& Find_exists_value( &User_opts, "logall", Value_sep ) ){
			Logall = 1;
		}
	}

	/* clean up statusfile information */
	if( Status_fd > 0 ){
		Errorcode = JABORT;
		FATAL(_("main: you opened the statusfile too early!!!"));
	}
	if( (s = Loweropts['s'-'a']) ) Statusfile = s;

	LOGMSG(_("main: using model '%s'"), Model_id?Model_id:_("DEFAULT"));

	/* if we have version request, print out the version */
	if( Debug || Find_flag_value(&Topts,"version",Value_sep)
		|| Find_flag_value(&Zopts,"version",Value_sep) ){
		Write_fd_str(2,_id);
	}

	/* DEBUG1("main: Debug level %d", Debug ); */
	if(DEBUGL2)Dump_line_list( "main: final Model information", &Model );
	if(DEBUGL3)Dump_line_list("Zopts",&Zopts);
	if(DEBUGL3)Dump_line_list("Topts",&Topts);

	(void)signal(SIGPIPE, SIG_IGN);
	(void)signal(SIGCHLD, SIG_DFL);

	/* initializes connection */
	if( Appsocket && !Device ){
		struct sockaddr sinaddr;
		memset(&sinaddr,0,sizeof(sinaddr));
		c = sizeof( sinaddr );
		if( (Device = Printcap_lp) ){
		} else if( fstat(1,&statb) == 0 && S_ISSOCK(statb.st_mode)
			&& getpeername( 1, &sinaddr, &c ) == 0 ){
			int port = 0;
			DEBUG1("main: family '%d'", sinaddr.sa_family );
			if( sinaddr.sa_family == AF_INET ){
				port = ((struct sockaddr_in *)&sinaddr)->sin_port;
#if defined(IPV6) 
			} else if( sinaddr.sa_family == AF_INET6 ){
				port = ((struct sockaddr_in6 * )&sinaddr)->sin6_port;
#endif
			} else {
				FATAL("main: bad protocol family '%d'", sinaddr.sa_family );
			}
			inet_ntop_sockaddr( &sinaddr, RemoteIPAddr, sizeof(RemoteIPAddr) );
			c = strlen(RemoteIPAddr);
			SNPRINTF(RemoteIPAddr+c, sizeof(RemoteIPAddr)-c) "%%%d",
				ntohs(port) );
			Device = RemoteIPAddr;
		} else {
			FATAL("main: appsocket and no dev= parameter, printcap, or network connection on fd 1");
		}
		DEBUG1("main: appsocket device now '%s'", Device );
	}
	if( !ISNULL(Device) ){
		close(1);
		Open_device( Device );
	} else if( Status ){
		Status = Fd_readable(1, &Poll_for_status);
	} 
	if( Poll_for_status ){
		/* handles the status case with the timeout */
		if( Dev_sleep <= 0 ) Dev_sleep = 100;
	}
	DEBUG1("main: poll for status %d, Dev_sleep %d msec",
		Poll_for_status, Dev_sleep);
	Set_block_io( 0 );
	Set_block_io( 1 );

	/* set the keepalive options */
	Set_keepalive(1);

	/* get the various options */
	if( (s = Find_str_value( &Model, "pjl_only", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_only, s, List_sep, 1, 0, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_only", &Pjl_only);
	}
	if( (s = Find_str_value( &Model, "pjl_except", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_except, s, List_sep, 1, 0, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_except", &Pjl_except);
	}
	if( (s = Find_str_value( &Model, "pjl_vars_set", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_vars_set, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_vars_set", &Pjl_vars_set);
	}
	if( (s = Find_str_value( &Model, "pjl_vars_except", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_vars_except, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_vars_except", &Pjl_vars_except);
	}
	if( (s = Find_str_value( &Model, "pjl_user_opts", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_user_opts, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_user_opts", &Pjl_user_opts);
	}
	if( (s = Find_str_value( &Model, "pcl_user_opts", Value_sep)) ){
		lowercase(s);
		Split( &Pcl_user_opts, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pcl_user_opts", &Pcl_user_opts);
	}
	if( (s = Find_str_value( &Model, "ps_user_opts", Value_sep)) ){
		lowercase(s);
		Split( &Ps_user_opts, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Ps_user_opts", &Ps_user_opts);
	}
	if( (s = Find_str_value( &Model, "pjl_error_codes", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_error_codes, s, "\n", 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_error_codes", &Pjl_error_codes);
	}
	if( (s = Find_str_value( &Model, "pjl_alert_codes", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_alert_codes, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_alert_codes", &Pjl_alert_codes);
	}
	if( (s = Find_str_value( &Model, "pjl_quiet_codes", Value_sep)) ){
		lowercase(s);
		Split( &Pjl_quiet_codes, s, List_sep, 1, Value_sep, 1, 1, 0 );
		if(DEBUGL4)Dump_line_list("main: Pjl_quiet_codes", &Pjl_quiet_codes);
	}

	do_pagecount = Check_pagecount(&pagecount_ps, &pagecount_pjl);
	/* now the real work */

	Process_job(do_pagecount, pagecount_ps, pagecount_pjl);

	/* cleanup and exit */
	Errorcode = 0;
	Set_block_io(1);
	cleanup(0);
	return(0);
}

void cleanup(int sig)
{
    DEBUG3("cleanup: Signal '%s', Errorcode %d", Sigstr(sig), Errorcode );
	Set_block_io( 1 );
	close(0);
	close(1);
	close(2);
    exit(Errorcode);
}

/*
 * getargs( int argc, char **argv )
 *  - get the options from the argument list
 */

void getargs( int argc, char **argv )
{
	int i, flag;
	char *arg, *s, m[2];

	/* LOGMSG("testing"); */
	DEBUG3("getargs: starting, debug %d", Debug);
	for( i = 1; i < argc; ++i ){
		arg = argv[i];
		if( *arg++ != '-' ) break;
		flag = *arg++;
		if( flag == '-' ){
			++i;
			break;
		}
		if( flag == 'c' ){
			 arg = "1";
		}
		if( *arg == 0 ){
			if( i < argc ){
				arg = argv[++i];
			} else {
				FATAL( "missing argument for flag '%c'", flag );
			}
		}
		/* we duplicate the strings */
		DEBUG3("getargs: flag '%c'='%s'", flag, arg);
		if( islower(flag) ){
			Loweropts[flag-'a'] = arg;
		} else if( isupper(flag) ){
			switch( flag ){
			case 'T': case 'Z':
				if( (s = Upperopts[flag-'A']) ){
					Upperopts[flag-'A'] = 
						safestrdup3(s,",",arg,__FILE__,__LINE__);
					if( s ) free(s);
				} else {
					Upperopts[flag-'A'] = safestrdup(arg,__FILE__,__LINE__);
				}
				break;
			default:
				Upperopts[flag-'A'] = arg;
				break;
			}
		}
	}
	if( i < argc ){
		Accountfile = argv[i];
	} else if( (s = Loweropts['a'-'a']) ){
		Accountfile = s;
	}
	DEBUG1("getargs: finished options");
	/* set the Topts and Zopts values */
	memset(m,sizeof(m),0);
	m[0] = m[1] = 0;
	for( i = 'a'; i <= 'z'; ++i ){
		m[0] = i;
		if( (s = Loweropts[i-'a']) ) Set_str_value(&Topts,m,s);
	}
	for( i = 'A'; i <= 'Z'; ++i ){
		m[0] = i;
		if( (s = Upperopts[i-'A']) ) Set_str_value(&Topts,m,s);
	}
	if( (s = Upperopts['F'-'A']) && cval(s) == 'o' ){
		OF_Mode = 1;
	}
	Fix_special_user_opts("Topts", &Topts, &Unsorted_Topts, Upperopts['T'-'A'] );
	Fix_special_user_opts("Zopts", &Zopts, &Unsorted_Zopts, Upperopts['Z'-'A'] );
}

void Fix_special_user_opts( char *name, struct line_list *opts, struct line_list *unsorted_opts,
	 char *line )
{
	struct line_list v;
	int i, append, c;
	char *key, *value, *s;

	DEBUG1("Fix_special_user_opts: %s - '%s'", name, line );
	Init_line_list( &v );
	Split( &v, line, ",", 0, 0, 0, 1, 0 );
	for( i = 0; i < v.count; ++i ){
		key = v.list[i];
		DEBUG1("Fix_special_user_opts: option '%s'", key );
		c = 0;
		if( (value = safestrpbrk( key, Value_sep )) ){
			c = *value;
			*value = 0;
		}
		lowercase(key);
		if( (append =  !strcasecmp( key, "font")) && c == '=' ){
			if( (s = Find_str_value(opts,key,Value_sep)) ){
				DEBUG1("Fix_special_user_opts: old '%s'='%s'", key, s );
				s = safestrdup3(s,",",value+1,__FILE__,__LINE__);
				DEBUG1("Fix_special_user_opts: new '%s'='%s'", key, s );
				Set_str_value(opts,key,s);
				if(s) free(s); s = 0;
			} else {
				DEBUG1("Fix_special_user_opts: init '%s'='%s'", key, value+1 );
				Set_str_value(opts,key,value+1);
			}
		}
		if( value ){
			*value = c;
		}
		if( !append ){
			Add_line_list(opts,key,Value_sep,1,1);
		}
		Add_line_list(unsorted_opts,key,0,0,0);
	}
	Free_line_list( &v );
	if(DEBUGL1){
		char msg[128];
		SNPRINTF(msg, sizeof(msg))  "Fix_special_user_opts - '%s' sorted", name );
		Dump_line_list(msg, opts );
		SNPRINTF(msg, sizeof(msg))  "Fix_special_user_opts - '%s' unsorted", name );
		Dump_line_list(msg, unsorted_opts );
	}
}

/*
 * Output buffer management
 *  Set up and put values into an output buffer for
 *  transmission at a later time
 */
void Init_outbuf()
{
	DEBUG1("Init_outbuf: Outbuf 0x%lx, Outmax %d, Outlen %d",
		(long)Outbuf, Outmax, Outlen );
	if( Outmax <= 0 ) Outmax = OUTBUFFER;
	if( Outbuf == 0 ) Outbuf = realloc_or_die( Outbuf, Outmax+1,__FILE__,__LINE__);
	Outlen = 0;
	Outbuf[0] = 0;
}

void Put_outbuf_str( char *s )
{
	if( s && *s ) Put_outbuf_len( s, strlen(s) );
}

void Put_outbuf_len( char *s, int len )
{
	DEBUG4("Put_outbuf_len: starting- Outbuf 0x%lx, Outmax %d, Outlen %d, len %d",
		(long)Outbuf, Outmax, Outlen, len );
	if( s == 0 || len <= 0 ) return;
	if( Outmax - Outlen <= len ){
		Outmax += ((OUTBUFFER + len)/1024)*1024;
		Outbuf = realloc_or_die( Outbuf, Outmax+1,__FILE__,__LINE__);
		DEBUG4("Put_outbuf_len: update- Outbuf 0x%lx, Outmax %d, Outlen %d, len %d",
		(long)Outbuf, Outmax, Outlen, len );
		if( !Outbuf ){
			Errorcode = JABORT;
			LOGERR_DIE( "Put_outbuf_len: realloc %d failed", len );
		}
	}
	memmove( Outbuf+Outlen, s, len );
	Outlen += len;
	Outbuf[Outlen] = 0;
}

/*
 * Input buffer management
 *  Set up and put values into an input buffer for
 *  scanning purposes
 */
void Init_inbuf()
{
	Inmax = LARGEBUFFER;
	Inbuf = realloc_or_die( Inbuf, Inmax+1,__FILE__,__LINE__);
	Inbuf[Inmax] = 0;
	Inlen = 0;
	Inbuf[0] = 0;
}

void Put_inbuf_len( char *str, int len )
{
	if( Inbuf == 0 || Inmax - Inlen <= (len+1) ){
		Inmax += ((LARGEBUFFER + len+1)/1024)*1024;
		Inbuf = realloc_or_die( Inbuf, Inmax+1,__FILE__,__LINE__);
		if( !Inbuf ){
			Errorcode = JABORT;
			LOGERR_DIE( _("Put_outbuf_len: realloc %d failed"), len );
		}
	}
	memmove( Inbuf+Inlen, str, len+1 );
	Inlen += len;
	Inbuf[Inlen] = 0;
	DEBUG4("Put_inbuf_len: buffer '%s'", Inbuf );
}

/*
 * void Get_inbuf_str()
 *  get a line from the input buffer and parse it for printer
 *  status
 */
void Get_inbuf_str(void)
{
	char *s, *t;
	/* DEBUG4("Get_inbuf_str: buffer '%s'", Inbuf); */
	Inbuf[Inlen] = 0;
	/* remove \r */
	for( s = Inbuf; (s = safestrchr( s, '\r' )); ){
		memmove( s, s+1, strlen(s)+1 );
	}
	for( s = Inbuf; (t = safestrpbrk( s, Line_ends )); s = t ){
		*t++ = 0;
		DEBUG4("Get_inbuf_str: found '%s'", s);
		Pr_status( s, 0 );
	}
	memmove( Inbuf, s, strlen(s)+1 );
	DEBUG4("Get_inbuf_str: final '%s'", Inbuf);
	Inlen = strlen(Inbuf);
}


/*
 * Printer Status Reporting
 * PJL:
 *  We extract info of the form:
 *  @PJL XXXX P1 P2 ...
 *  V1
 *  V2
 *  0
 *    we then create a string 
 *  P1.P2.P3=xxxx
 *  - we put this into the 'current status' array
 * PostScript:
 *  We extract info of the form:
 *  %%[ key: value value key: value value ]%%
 *    we then create strings
 *     key=value value
 *  - we put this into the 'current status' array
 *
 */

void Pr_status( char *str, int monitor_status )
{
	char *s, *t;
	int c, i, pjl_line, found;
	char *ps_str;
	struct line_list l;
	static char *pjlvar, *eq_line;
	char *value;

	c = pjl_line = found = 0;
	ps_str = value = eq_line = 0;
	while( str && isspace( cval(str) ) ) ++str;
	DEBUG4("Pr_status: start str '%s', pjlvar '%s'", str, pjlvar);
	/* if(DEBUGL4)Dump_line_list("Pr_status - before Devstatus", &Devstatus); */

	Init_line_list(&l);
	if( Logall && str ){
		LOGMSG( _("Pr_status: printer status '%s'"), str );
	}
	/* if the previous line was a PJL line,  then the
	 * last entry may be a variable name and the next
	 * line a value for it.
	 */
	if( (ps_str = safestrstr( str,"%%[" )) ){
		ps_str += 3;
	}
	if( (s = safestrpbrk(str,Whitespace)) ){
		c = cval(s);
		*s = 0;
	}
	pjl_line = (safestrstr( str, "@PJL") != 0 );
	eq_line = safestrchr( str, '=' );
	if( s ) *s = c;

	/* now we check to see if we have a value for pjl variable */
	if( pjlvar && (pjl_line || eq_line || ps_str ) ){
		free(pjlvar); pjlvar = 0;
	}
	if( pjlvar ){
		DEBUG4("Pr_status: setting PJL var '%s' to '%s'", pjlvar, str);
		Set_str_value( &Devstatus, pjlvar, str );
		free(pjlvar); pjlvar = 0;
		goto done;
	}

	/* now we check for xx=value entries */
	if( eq_line ){
		*eq_line = 0;
		lowercase(str);
		*eq_line = '=';
		Check_device_status(str);
		DEBUG4("Pr_status: setting '%s'", str);
		Add_line_list( &Devstatus, str, Value_sep, 1, 1 );
		goto done;
	}
	if( pjl_line ){
		DEBUG4("Pr_status: doing PJL status on '%s'", str);
		if( (s = safestrstr( str, " ECHO ") ) ){
			/* we have the echo value */
			Set_str_value(&Devstatus,"echo",s+6);
			DEBUG4("Pr_status: found echo '%s'", s+6 );
		} else {
			Split(&l, str, Whitespace, 0, 0, 0, 1, 0 );
			pjlvar = safestrdup(l.list[l.count-1],__FILE__,__LINE__);
			Free_line_list( &l );
			lowercase(pjlvar);
			DEBUG4("Pr_status: PJL var '%s'", pjlvar);
		}
	} else if( monitor_status || ps_str ){
		/* we do Postscript status */
		if( !ps_str ) ps_str = str;
		Split(&l,ps_str,Whitespace,0,0,0,1,0 );
		DEBUG4("Pr_status: PostScript info '%s'", ps_str );
		if(DEBUGL4)Dump_line_list("Pr_status: PostScript", &l);
		for( i = 0; i < l.count; ++i ){
			s = l.list[i];
			found = 0;
			t = 0;
			DEBUG4("Pr_status: list [%d]='%s', value %s", i, s, value );
			if( safestrstr( s, "]%%" ) || safestrstr( s, "%%[" )
				|| (found = (strlen(s) > 1
						&& (t = safestrchr(s,':')) && t[1] == 0)) ){
				if( value ){
					DEBUG4("Pr_status: found ps status '%s'", value );
					Check_device_status(value);
					Add_line_list( &Devstatus, value, Value_sep, 1, 1 );
					free( value ); value = 0;
				}
				c = 0;
			}
			if( found ){
				*t = '=';
				value = safestrdup(s,__FILE__,__LINE__);
				lowercase(value);
				c = 0;
			} else if( value ){
				t = value;
				value = safestrdup3(value,c?" ":"",s,__FILE__,__LINE__);
				free(t);
				c = 1;
			}
		}
		if( value ){
			DEBUG5("Pr_status: found ps status '%s'", value );
			Check_device_status(value);
			Add_line_list( &Devstatus, value, Value_sep, 1, 1 );
			free( value ); value = 0;
		}
		Free_line_list( &l );
	}
  done:
	/* if(DEBUGL4)Dump_line_list("Pr_status - Devstatus", &Devstatus); */
	;
}

/*
 * Check_device_status( char *line )
 *  - the line has to have the form DEVICE="nnn"
 *  We log the message IF it is different from the last one
 */

void Check_device_status( char *line )
{
	char *old, *value, *handler = 0;
	int c, i, alert, fd, found;
	char msg[SMALLBUFFER];

	DEBUG4("Check_device_status: '%s'", line );
	if( (value = safestrpbrk( line, Value_sep )) ){
		c = *value;
		*value++ = 0;
		DEBUG4("Check_device_status: key '%s', value '%s'", line, value );
		if( !(old = Find_str_value( &Devstatus, line, Value_sep ))
			|| strcmp(old,value) ){
			if( !strcasecmp(line,"code") ){
				DEBUG4("Check_device_status: CODE '%s'", value );
				/* we now have to check to see if we log this */
				found = 0;
				for( i = 0; !found && i < Pjl_quiet_codes.count; ++i ){
					found = !Globmatch( Pjl_quiet_codes.list[i], value );
				}
				DEBUG4("Check_device_status: CODE '%s' found %d", value, found );
				if( !found ){
					/* ok, we log it */
					Check_code( value, msg, sizeof(msg) );
					alert = 0;
					for( i = 0; !alert && i < Pjl_alert_codes.count; ++i ){
						alert = !Globmatch(Pjl_alert_codes.list[i],value);
					}
					if( alert ){
						handler = Find_str_value( &Model,"pjl_alert_handler",
								Value_sep );
						handler = Fix_option_str( handler, 0, 1, 1 );
					}
					DEBUG4("Check_device_code: alert %d, handler '%s'",alert,handler);
					LOGMSG(_("Check_device_status: code = %s, '%s'%s"),
						value, msg, alert?", ALERT OPERATOR":"" );
					if( handler ){
						fd = Make_tempfile();
						if( Write_fd_str( fd, msg ) < 0 || Write_fd_str(fd, "\n" ) < 0 ){
							LOGERR_DIE( _("Check_device_status: write(fd %d) failed"), fd );
						}
						if( lseek(fd,0,SEEK_SET) == -1 ){
							Errorcode = JABORT;
							LOGERR_DIE(_("Check_device_status: lseek failed"));
						}
						Filter_file( handler, _("ALERT"), fd, -1, 0, 0 );
						close(fd);
					}
					if( handler ) free( handler ); handler = 0;
				}
			} else if(
				safestrstr(line,"status")
				|| safestrstr(line,"error")
				|| safestrstr(line,"offending")
				|| safestrstr(line,_("status"))
				|| safestrstr(line,_("error"))
				|| safestrstr(line,_("offending"))
				){
				DEBUG4("Check_device_status: old status '%s', new '%s'",
					old, value );
				LOGMSG("Check_device_status: %s = '%s'", line, value );
			}
		}
		value[-1] = c;
	}
}


/*
 * void Initialize_parms( struct line_list *list, struct keyvalue *valuelist )
 *  Initialize variables from values in a parameter list
 *  This list has entries of the form key=value
 *  If the variable is a flag,  then entry of the form 'key' will set
 *    it,  and key@ will clear it.
 *
 *  list = key list, i.e. - strings of form key=value
 *  count = number of entries in key list
 *  valuelist = variables and keys to use to set them
 */

void Initialize_parms( struct line_list *list, struct keyvalue *valuelist )
{
	struct keyvalue *v;
	char *arg, *convert;
	int n;
	if( valuelist == 0 ) return;
	for( v = valuelist; v->key; ++v ){
		if( v->key[0] == 0 ) continue;
		if( Find_first_key( list, v->key, Value_sep, 0) ) continue;
		arg = Find_exists_value( list, v->key, Value_sep );
		DEBUG4("Initialize_parms: key '%s'='%s'", v->key, arg );
		if( arg ) switch( v->kind ){
		case STRV: *(char **)v->var = arg; break;
		case INTV:
		case FLGV:
			convert = arg;
			n = strtol( arg, &convert, 10 );
			if( convert != arg ){
				*(int *)v->var = n;
			} else {
				*(int *)v->var = (
				!strcasecmp( arg, "yes" )
				|| !strcasecmp( arg, "on" )
				|| !strcasecmp( arg, _("yes") )
				|| !strcasecmp( arg, _("on") )
				);
			}
			break;
		}
	}
}

/*
 * Dump_parms( char *title, struct keyvalue *v )
 *  Dump the names, config file tags, and current values of variables
 *   in a value list.
 */
void Dump_parms( char *title, struct keyvalue *v )
{
	LOGDEBUG( "Dump_parms: %s", title );
	for( v = Valuelist; v->key; ++v ){
		switch( v->kind ){
		case STRV:
			LOGDEBUG( " '%s' (%s) STRV = '%s'", v->varname, v->key, *(char **)v->var ); break;
		case INTV:
			LOGDEBUG( " '%s' (%s) INTV = '%d'", v->varname, v->key, *(int *)v->var ); break;
		case FLGV:
			LOGDEBUG( " '%s' (%s) FLGV = '%d'", v->varname, v->key, *(int *)v->var ); break;
		}
	}
}

/*
 * Process the job  - we do this for both IF and OF modes 
 */



void Process_job( int do_pagecount, int pagecount_ps, int pagecount_pjl )
{
	struct stat statb;
	int in_fd, out_fd, tempfd, startpagecounter, suspend;

	Init_outbuf();

	startpagecounter = 0;

	Pjl_displayname();
	LOGMSG( _("Process_job: setting up printer"));
	in_fd = out_fd = tempfd = -1;
	if( OF_Mode ){
		if( (in_fd=dup(0)) == -1){
			Errorcode = JFAIL;
			LOGERR_DIE( _("Process_job: dup(0) failed") );
		}
		if( !Appsocket && (out_fd=dup(1)) == -1){
			Errorcode = JFAIL;
			LOGERR_DIE( _("Process_job: dup(1) failed") );
		}
		tempfd = Make_tempfile();
		LOGMSG( _("Process_job: starting OF mode passthrough") );
		Start_of_job(&startpagecounter, do_pagecount, pagecount_ps,
			pagecount_pjl, 0 );
		do{
			if( lseek(tempfd,0,SEEK_SET) == -1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Process_job: lseek of tempfd failed"));
			}
			if( ftruncate(tempfd,0) == -1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Process_job: ftruncate of tempfd failed"));
			}
			/* see if we need to suspend */
			if( dup2(tempfd,1) == -1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Process_job: tempfd dup2(%d,1) failed"), tempfd);
			}
			suspend = Process_OF_mode();
			if( fstat(tempfd,&statb) == -1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Process_job: stat failed"));
			}
			/* anything to write */
			if( statb.st_size ){
				double f = statb.st_size;
				if( lseek(tempfd,0,SEEK_SET) == -1 ){
					Errorcode = JABORT;
					LOGERR_DIE(_("Process_job: lseek failed"));
				}
				if( dup2(tempfd,0) == -1 ){
					Errorcode = JFAIL;
					LOGERR_DIE( _("Process_job: tempfd dup2() failed") );
				}
				if( Appsocket ){
					Open_device( Device );
					DEBUG1("Process_job: Appsocket device open done");
				} else if( dup2(out_fd,1) == -1 ){
					Errorcode = JFAIL;
					LOGERR_DIE( _("Process_job: outfd dup2(%d,1) failed"), out_fd );
				}
				Start_of_job(0,0,0,0,1);
				LOGMSG( _("Process_job: sending %0.0f bytes of OF input"), f );
				Send_job();
				End_of_job(0,0,0,0,Wait_for_banner,1,1);
			}
			if( suspend ){
				if( Appsocket ){
					int len;
					LOGMSG(_("Process_job: using shutdown on appsocket connection"));
					shutdown(1,1);
					do{
						len = Read_status_timeout(0);
					} while( len == 0 );
					/* we shut down the socket connection */
					DEBUG1("Process_job: Appsocket device close");
					close(1);
				} else {
					Set_block_io(out_fd);
				}
				LOGMSG( _("Process_job: OF process suspending") );
				if( Status_fd > 0 ){
					close( Status_fd );
					Status_fd = -2;
				}
				kill(getpid(), SIGSTOP);
				LOGMSG( _("Process_job: OF process running") );
			}
			if( dup2(in_fd,0) == -1 ){
				Errorcode = JFAIL;
				LOGERR_DIE( _("Process_job: infd dup2(%d,1) failed"), in_fd );
			}
		} while( suspend );
		if( !Appsocket && dup2(out_fd,1) == -1 ){
			Errorcode = JFAIL;
			LOGERR_DIE( _("Process_job: outfd dup2(%d,1) failed"), out_fd );
		}
		close(tempfd);
		close(in_fd);
		close(out_fd);
		End_of_job(&startpagecounter, do_pagecount, pagecount_ps,
			pagecount_pjl, 1, 0,0);
		LOGMSG( _("Process_job: ending OF mode passthrough") );
	} else {
		Start_of_job(&startpagecounter, do_pagecount, pagecount_ps,
			pagecount_pjl, 0 );
		LOGMSG( _("Process_job: sending job file") );
		Send_job();
		LOGMSG( _("Process_job: sent job file") );
		End_of_job(&startpagecounter, do_pagecount, pagecount_ps,
			pagecount_pjl, 1, 0, 0 );
	}
	LOGMSG( _("Process_job: done") );
}


void Start_of_job( int *startpagecounter, int do_pagecount, int pagecount_ps,
	int pagecount_pjl, int nested_job )
{
	char *s;
	struct line_list l;
	struct stat statb;

	Init_line_list(&l);
	DEBUG1("Start_of_job: do_pagecount %d, ps %d, pjl %d",
		do_pagecount, pagecount_ps, pagecount_pjl );
	if( Appsocket && fstat(1,&statb) == -1 ){
		Open_device( Device );
		DEBUG1("Start_of_job: Appsocket device open done");
	}
	if( fstat(1,&statb) == -1 ){
		Errorcode = JABORT;
		FATAL( _("Start_of_job: fd 1 not open") );
	}
	Set_block_io(1);
	if( Status ){
		Set_str_value(&Devstatus,"pagecount",0);
		Do_sync(Sync_timeout, Sync_interval, pagecount_pjl && do_pagecount );
		if( do_pagecount && Pagecount_start && startpagecounter ){
			*startpagecounter = Do_pagecount(Pagecount_timeout,
				Pagecount_interval, Pagecount_poll,
				pagecount_ps, pagecount_pjl);
			/* if we have Appsocket then we need to reopen the connection */
		}
		if( Appsocket ){
			DEBUG1("Start_of_job: Appsocket device close");
			close(1);
			Open_device( Device );
			DEBUG1("Start_of_job: Appsocket device open done");
		}
	}
	if( startpagecounter ){
		Do_accounting(1, 0, *startpagecounter, 0 );
	}

	Init_outbuf();
	DEBUG1("Start_of_job: doing 'init'");
	if( !Find_first_key( &Model, "init", Value_sep, 0) ){
		s  = Find_str_value( &Model, "init", Value_sep);
		DEBUG1("Start_of_job: 'init'='%s'", s);
		Split( &l, s, List_sep, 0, 0, 0, 1, 0 );
		Resolve_list( "", &l, Put_pcl );
		Free_line_list( &l );
	}
	if( Pjl ){
		DEBUG1("Start_of_job: doing pjl");
		Put_outbuf_str( PJL_UEL_str );
		Put_outbuf_str( PJL_str );
		if( !nested_job) Pjl_job();
		Pjl_console_msg(1);
		if( (s  = Find_str_value( &Model, "pjl_init", Value_sep)) ){
			DEBUG1("Start_of_job: 'pjl_init'='%s'", s);
			Split( &l, s, List_sep, 0, 0, 0, 1, 0 );
			Resolve_list( "pjl_", &l, Put_pjl );
			Free_line_list( &l );
		}
		if( !OF_Mode ){
			DEBUG1("Start_of_job: 'pjl' and Zopts");
			Resolve_user_opts( "pjl_", &Pjl_user_opts, &Unsorted_Zopts, Put_pjl );
		}
		DEBUG1("Start_of_job: 'pjl' and Topts");
		Resolve_user_opts( "pjl_", &Pjl_user_opts, &Unsorted_Topts, Put_pjl );
	}
	if( Write_read_timeout( Outlen, Outbuf, Job_timeout ) ){
		Errorcode = JFAIL;
		FATAL(_("Start_of_job: timeout"));
	}
	Init_outbuf();
	Free_line_list(&l);
}


void End_of_job( int *startpagecounter, int do_pagecount, int pagecount_ps,
	int pagecount_pjl, int wait_for_end, int nested_job, int banner_page )
{
	char *s;
	struct line_list l;
	int endpagecounter = 0;
	struct stat statb;

	Init_line_list(&l);
	Init_outbuf();
	if( fstat(1,&statb) == -1 ){
		if( Appsocket ){
			Open_device( Device );
			DEBUG1("End_of_job: Appsocket device open done");
		} else {
			Errorcode = JFAIL;
			LOGERR_DIE(_("End_of_job: FD 1 IS CLOSED AND ifhp DID NOT CLOSE IT"));
		}
	}
	Set_block_io(1);
	if( Pjl ){
		DEBUG1("End_of_job: doing pjl at end");
		Put_outbuf_str( PJL_UEL_str );
		Put_outbuf_str( PJL_str );
		Pjl_console_msg(1);
		if( !nested_job ) Pjl_eoj(Jobname);
		if( !Find_first_key( &Model, "pjl_term", Value_sep, 0) ){
			s  = Find_str_value( &Model, "pjl_term", Value_sep);
			DEBUG1("End_of_job: 'pjl_term'='%s'", s);
			Split( &l, s, List_sep, 0, 0, 0, 1, 0 );
			Resolve_list( "pjl_", &l, Put_pjl );
			Free_line_list( &l );
		}
	}
	if( !Find_first_key( &Model, "term", Value_sep, 0) ){
		s  = Find_str_value( &Model, "term", Value_sep);
		DEBUG1("End_of_job: 'term'='%s'", s);
		Split( &l, s, List_sep, 0, 0, 0, 1, 0 );
		Resolve_list( "", &l, Put_pcl );
		Free_line_list( &l );
	}
	if( Write_read_timeout( Outlen, Outbuf, Job_timeout ) ){
		Errorcode = JFAIL;
		FATAL(_("End_of_job: timeout"));
	}
	Init_outbuf();
	DEBUG1("End_of_job: end sync and pagecount, do_pagecount %d, ps %d, pjl %d",
		do_pagecount, pagecount_ps, pagecount_pjl );

	/*
	 * Some badly bent or totally broken appsocket printers MUST
	 * have a shutdown rather than a close done on the connection,
	 * and the connection MUST stay open and MUST be read UNTIL
	 * the printer is finished.  This is really really maddening,
	 * but that is the way it is done.
	 *
	 * if we are doing Appsocket
	 *   and we have shutdown_appsocket
	 * THE we shutdown the Appsocket and read from it until EOF
	 *   We discard the data if Status@
	 */
	if( Appsocket ){
		int len;
		shutdown(1,1);
		do{
			len = Read_status_timeout( 0 );
		} while( len == 0 );
	} else if( Status && wait_for_end ){
		/* we wait for the end normally */
		Do_waitend(Job_timeout, Waitend_interval, Waitend_ctrl_t_interval, banner_page );
	}

	/* now get the pagecounter value */
	Set_str_value(&Devstatus,"pagecount",0);
	if( Status && do_pagecount && Pagecount_end && startpagecounter ){
		endpagecounter = Do_pagecount(Job_timeout, Pagecount_interval,
			Pagecount_poll, pagecount_ps, pagecount_pjl);
	}

	if( Pjl && Pjl_console ){
		if( Appsocket ){
			/* we shut down the socket connection */
			DEBUG1("End_of_job: Appsocket device close");
			close(1);
			DEBUG1("End_of_job: Appsocket device open done");
			Open_device( Device );
		}
		Init_outbuf();
		Pjl_console_msg(0);
		DEBUG1("End_of_job: clearing console at end");
		Put_outbuf_str( PJL_UEL_str );
		Write_read_timeout( Outlen, Outbuf, Job_timeout );
		Init_outbuf();
	}

	if( startpagecounter ){
		time_t current_t;
		int elapsed, n;

		elapsed = n = 0;
		time( &current_t );
		elapsed = current_t - Start_time;
		if( Pagecount_start && Pagecount_end ){
			n = endpagecounter - *startpagecounter;
		}
		Do_accounting(0, elapsed, endpagecounter, n );
	}

	if( Appsocket ){
		/* we shut down the socket connection */
		DEBUG1("End_of_job: Appsocket device close");
		close(1);
	}
	Free_line_list(&l);
}


/*
 * Put_pjl( char *s )
 *  write pjl out to the output buffer
 *  check to make sure that the line is PJL code only
 *  and that the PJL option is in a list
 */

void Put_pjl( char *s )
{
	struct line_list l, wl;
	int i;

	Init_line_list(&l);
	Init_line_list(&wl);

	DEBUG4("Put_pjl: orig '%s'", s );
	if( s == 0 || *s == 0 ) return;
	s = Fix_option_str( s, 0, 1, 0 );
	Strip_leading_spaces(&s);
	Split( &l, s, "\n", 0, 0, 0, 1, 0 );
	if( s ){ free(s); s = 0; }
	for( i = 0; i < l.count; ++i ){
		/* check for valid PJL */
		Split( &wl, l.list[i], Whitespace, 0, 0, 0, 0, 0 );
		if( wl.count == 0 || strcmp("@PJL", wl.list[0]) ){
			continue;
		}
		/* key must be in PJL_ONLY list if supplied, and NOT in
		 * PJL_ACCEPT
		 */
		s = wl.list[1];
		lowercase(s);
		DEBUG4("Put_pjl: trying '%s'", s );
		if( wl.count > 1 &&
			(  Find_first_key( &Pjl_only, s, 0, 0)
			||  !Find_first_key( &Pjl_except, s, 0, 0)
			) ){
			continue;
		}
		DEBUG4("Put_pjl: '%s'", l.list[i] );
		Put_outbuf_str( l.list[i] );
		Put_outbuf_str( "\n" );
		Free_line_list( &wl );
	}
	Free_line_list(&l);
}

void Put_pcl( char *s )
{
	DEBUG4("Put_pcl: orig '%s'", s );
#if defined(TRIMPCL)
	if( s ){
		char *t;
		while( isspace(cval(s)) ) ++s;
		t = s + strlen(s);
		while( t > s && isspace(cval(t-1)) ) --t;
		*t = 0;
	}
	DEBUG4("Put_pcl: trimmed '%s'", s );
	s = Fix_option_str( s, 0, 0, 0 );
#else
	s = Fix_option_str( s, 1, 1, 0 );
#endif
	DEBUG4("Put_pcl: final '%s'", s );
	Put_outbuf_str( s );
	if(s)free(s);s=0;
}


void Put_ps( char *s )
{
	DEBUG4("Put_ps: orig '%s'", s );
	s = Fix_option_str( s, 0, 0, 0 );
	DEBUG4("Put_ps: final '%s'", s );
	Put_outbuf_str( s );
	if(s)free(s);s=0;
	Put_outbuf_str( "\n" );
}


void Put_fixed( char *s )
{
	DEBUG4("Put_fixed: orig '%s'", s );
	s = Fix_option_str( s, 0, 0, 0 );
	DEBUG4("Put_fixed: final '%s'", s );
	Put_outbuf_str( s );
	if(s)free(s);s=0;
}


/*
 * Set_non_block_io(fd)
 * Set_block_io(fd)
 *  Set blocking or non-blocking IO
 *  Dies if unsuccessful
 * Get_nonblock_io(fd)
 *  Returns O_NONBLOCK flag value
 */

int Get_nonblock_io( int fd )
{
	int mask;
	/* we set IO to non-blocking on fd */

	mask = fcntl( fd, F_GETFL, 0 );
	if( mask != -1 ) mask &= O_NONBLOCK;
	DEBUG4("Get_nonblock_io: fd %d, current flags 0x%x, O_NONBLOCK=0x%x",fd, mask, O_NONBLOCK);
	return( mask );
}

void Set_nonblock_io( int fd )
{
	int omask;
	/* we set IO to non-blocking on fd */

	if( (omask = fcntl( fd, F_GETFL, 0 ) ) == -1 ) return;
	omask |= O_NONBLOCK;
	fcntl( fd, F_SETFL, omask );
}

void Set_block_io( int fd )
{
	int omask;
	/* we set IO to blocking on fd */

	if( (omask = fcntl( fd, F_GETFL, 0 ) ) == -1 ) return;
	omask &= ~O_NONBLOCK;
	fcntl( fd, F_SETFL, omask );
}


/**************************************************************
 * 
 * signal handling:
 * SIGALRM should be the only signal that terminates system calls;
 * all other signals should NOT terminate them.
 * This signal() emulation function attepts to do just that.
 * (Derived from Advanced Programming in the UNIX Environment, Stevens, 1992)
 *
 **************************************************************/


/* solaris 2.3 note: don't compile this with "gcc -ansi -pedantic";
 * due to a bug in the header file, struct sigaction doesn't
 * get declared. :(
 */

/* plp_signal will set flags so that signal handlers will continue
 * note that in Solaris,  you MUST reinstall the
 * signal hanlders in the signal handler!  The default action is
 * to try to restart the system call - note that the code should
 * be written so that you check for error returns, and continue
 * so this is merely a convenience.
 */

plp_sigfunc_t plp_signal (int signo, plp_sigfunc_t func)
{
#ifdef HAVE_SIGACTION
	struct sigaction act, oact;

	act.sa_handler = func;
	(void) sigemptyset (&act.sa_mask);
	act.sa_flags = 0;
# ifdef SA_RESTART
	act.sa_flags |= SA_RESTART;             /* SVR4, 4.3+BSD */
# endif
	if (sigaction (signo, &act, &oact) < 0) {
		return (SIG_ERR);
	}
	return (plp_sigfunc_t) oact.sa_handler;
#else
	/* sigaction is not supported. Just set the signals. */
	return (plp_sigfunc_t)signal (signo, func); 
#endif
}

/* plp_signal_break is similar to plp_signal,  but will cause
 * TERMINATION of a system call if possible.  This allows
 * you to force a signal to cause termination of a system
 * wait or other action.
 */

plp_sigfunc_t plp_signal_break (int signo, plp_sigfunc_t func)
{
#ifdef HAVE_SIGACTION
	struct sigaction act, oact;

	act.sa_handler = func;
	(void) sigemptyset (&act.sa_mask);
	act.sa_flags = 0;
# ifdef SA_INTERRUPT
	act.sa_flags |= SA_INTERRUPT;            /* SunOS */
# endif
	if (sigaction (signo, &act, &oact) < 0) {
		return (SIG_ERR);
	}
	return (plp_sigfunc_t) oact.sa_handler;
#else
	/* sigaction is not supported. Just set the signals. */
	return (plp_sigfunc_t)signal (signo, func); 
#endif
}

/**************************************************************/

void plp_block_all_signals ( plp_block_mask *oblock )
{
#ifdef HAVE_SIGPROCMASK
	sigset_t block;

	(void) sigfillset (&block); /* block all signals */
	if (sigprocmask (SIG_SETMASK, &block, oblock) < 0)
		LOGERR_DIE( _("plp_block_all_signals: sigprocmask failed"));
#else
	*oblock = sigblock( ~0 ); /* block all signals */
#endif
}


void plp_unblock_all_signals ( plp_block_mask *oblock )
{
#ifdef HAVE_SIGPROCMASK
	sigset_t block;

	(void) sigemptyset (&block); /* block all signals */
	if (sigprocmask (SIG_SETMASK, &block, oblock) < 0)
		LOGERR_DIE(_("plp_unblock_all_signals: sigprocmask failed"));
#else
	*oblock = sigblock( 0 ); /* unblock all signals */
#endif
}

void plp_set_signal_mask ( plp_block_mask *in, plp_block_mask *out )
{
#ifdef HAVE_SIGPROCMASK
	if (sigprocmask (SIG_SETMASK, in, out ) < 0)
		LOGERR_DIE(_("plp_set_signal_mask: sigprocmask failed"));
#else
	if( out ){
		*out = sigblock( *in ); /* block all signals */
	} else {
	 	(void)sigblock( *in ); /* block all signals */
	}
#endif
}

void plp_unblock_one_signal ( int sig, plp_block_mask *oblock )
{
#ifdef HAVE_SIGPROCMASK
	sigset_t block;

	(void) sigemptyset (&block); /* clear out signals */
	(void) sigaddset (&block, sig ); /* clear out signals */
	if (sigprocmask (SIG_UNBLOCK, &block, oblock ) < 0)
		LOGERR_DIE(_("plp_unblock_one_signal: sigprocmask failed"));
#else
	*oblock = sigblock( 0 );
	(void) sigsetmask (*oblock & ~ sigmask(sig) );
#endif
}

void plp_block_one_signal( int sig, plp_block_mask *oblock )
{
#ifdef HAVE_SIGPROCMASK
	sigset_t block;

	(void) sigemptyset (&block); /* clear out signals */
	(void) sigaddset (&block, sig ); /* clear out signals */
	if (sigprocmask (SIG_BLOCK, &block, oblock ) < 0)
		LOGERR_DIE(_("plp_block_one_signal: sigprocmask failed"));
#else
	*oblock = sigblock( sigmask( sig ) );
#endif
}

void plp_sigpause( void )
{
#ifdef HAVE_SIGPROCMASK
	sigset_t block;
	(void) sigemptyset (&block); /* clear out signals */
	(void) sigsuspend( &block );
#else
	(void)sigpause( 0 );
#endif
}


/***************************************************************************
 * Set up alarms so LPRng doesn't hang forever during transfers.
 ***************************************************************************/

/*
 * timeout_alarm
 *  When we get the alarm,  we close the file descriptor (if any)
 *  we are working with.  When we next do an option, it will fail
 *  Note that this will cause any ongoing read/write operation to fail
 * We then to a longjmp to the routine, returning a non-zero value
 * We set an alarm using:
 *
 * if( (setjmp(Timeout_env)==0 && Set_timeout_alarm(t,s)) ){
 *   timeout dependent stuff
 * }
 * Clear_alarm
 * We define the Set_timeout macro as:
 *  #define Set_timeout(t,s) (setjmp(Timeout_env)==0 && Set_timeout_alarm(t,s))
 */

 static plp_signal_t timeout_alarm (int sig)
{
	Alarm_timed_out = 1;
	signal( SIGALRM, SIG_IGN );
	errno = EINTR;
#if defined(HAVE_SIGLONGJMP)
	siglongjmp(Timeout_env,1);
#else
	longjmp(Timeout_env,1);
#endif
}


 static plp_signal_t timeout_break (int sig)
{
	Alarm_timed_out = 1;
	signal( SIGALRM, SIG_IGN );
}


/***************************************************************************
 * Set_timeout( int timeout, int *socket )
 *  Set up a timeout to occur; note that you can call this
 *   routine several times without problems,  but you must call the
 *   Clear_timeout routine sooner or later to reset the timeout function.
 *  A timeout value of 0 never times out
 * Clear_alarm()
 *  Turns off the timeout alarm
 ***************************************************************************/
void Set_timeout_signal_handler( int timeout, plp_sigfunc_t handler )
{
	int err = errno;
	sigset_t oblock;

	alarm(0);
	signal(SIGALRM, SIG_IGN);
	plp_unblock_one_signal( SIGALRM, &oblock );
	Alarm_timed_out = 0;
	Timeout_pending = 0;

	if( timeout > 0 ){
		Timeout_pending = timeout;
		plp_signal_break(SIGALRM, handler);
		alarm (timeout);
	}
	errno = err;
}


void Set_timeout_alarm( int timeout )
{
	Set_timeout_signal_handler( timeout, timeout_alarm );
}

void Set_timeout_break( int timeout )
{
	Set_timeout_signal_handler( timeout, timeout_break );
}

void Clear_timeout( void )
{
	int err = errno;

	signal( SIGALRM, SIG_IGN );
	alarm(0);
	Timeout_pending = 0;
	errno = err;
}

/*
 * Write_fd_len_timeout
 *  write the buffer of length len to file descriptor fd
 *  within timeout seconds (timeout < 0 means no timeout)
 *   returns:
 *    = -1  error
 *    = -2  timeout
 *    0   - no error
 */

int Write_fd_len_timeout( int timeout, int fd, const char *msg, int len )
{
	int i;
	Set_timeout_break( timeout );
	i = write( fd, msg, len );
	Clear_timeout();
	if( i == -1 && Alarm_timed_out ){
		i = -2;
	}
	return( i );
}

/*
 * Write_read_timeout( int readfd, int *flag,
 *	int writefd, char *buffer, int len, int timeout )
 *  Write the contents of a buffer to the file descriptor
 *   and then optionally read the status
 *  char *buffer, int len: buffer and number of bytes
 *  int timeout:  timeout in seconds
 *
 *   Returns:
 *    0 = success
 *    = -1 error
 *    = -2 timeout
 */

int Write_read_timeout( int len, char *buffer, int timeout )
{
	time_t start_t, current_t;
	int elapsed, m, left, err, n;
	struct timeval timeval, *tp;
    fd_set readfds, writefds; /* for select */

	DEBUG4( "Write_read_timeout: Status %d, Poll_for_status %d, write(len %d) timeout %d",
		Status, Poll_for_status, len, timeout );

	time( &start_t );

	/* do we do only a read ? */
	if( len <= 0 ) return 0;
	if( Status == 0 ){
		DEBUG4( "Write_read_timeout: blocking write len %d", len );
		m = Write_fd_len_timeout( timeout, 1, buffer, len );
		DEBUG4( "Write_read_timeout: block write status %d", m );
		len = 0;
		if( m < 0 ){
			LOGERR( _("Write_read_timeout: write error or timeout"));
			len = m;
		}
	} else if( Poll_for_status ){
		DEBUG4( "Write_read_timeout: poll write len %d", len );
		m = Read_status_timeout(-1);
		DEBUG1("Write_read_timeout: poll Read_status_timeout returned %d", m );
		while( len > 0 ){
			left = 0;
			if( timeout > 0 ){
				time( &current_t );
				elapsed = current_t - start_t;
				left = timeout - elapsed;
				DEBUG4("Write_read_timeout: poll timeout left %d", left );
				if( left <= 0 ){
					len = -2;
					break;
				}
			}
			DEBUG1("Write_read_timeout: poll writing %d", len );
			m = Write_fd_len_timeout( left, 1, buffer, len );
			DEBUG4( "Write_read_timeout: poll write status %d", m );
			if( m < 0 ){
				LOGERR( _("Write_read_timeout: poll write error"));
			} else {
				len -= m;
				buffer += m;
			}
			m = Read_status_timeout(-1);
			DEBUG1("Write_read_timeout: poll Read_status_timeout returned %d", m );
		}
	} else while( len > 0 ){
		if( timeout > 0 ){
			time( &current_t );
			elapsed = current_t - start_t;
			left = timeout - elapsed;
			DEBUG4("Write_read_timeout: timeout left %d", left );
			if( left <= 0 ){
				len = -2;
				break;
			}
			memset( &timeval, 0, sizeof(timeval) );
			timeval.tv_sec = left;
			tp = &timeval;
		} else {
			tp = 0;
		}
		FD_ZERO( &writefds );
		FD_ZERO( &readfds );
		m = 0;
		if( len > 0 ){
			FD_SET( 1, &writefds );
			DEBUG4("Write_read_timeout: setting write fd %d", 1);
			if( m <= 1 ) m = 2;
		}
		if( Status >= 0 ){
			FD_SET( 1, &readfds );
			DEBUG4("Write_read_timeout: setting read fd %d", 1 );
			if( m <= 1 ) m = 2;
		}
		DEBUG4("Write_read_timeout: starting rw select - max %d", m );
		errno = 0;
		m = select( m,
			FD_SET_FIX((fd_set *))&readfds,
			FD_SET_FIX((fd_set *))&writefds,
			FD_SET_FIX((fd_set *))0, tp );
		err = errno;
		DEBUG4("Write_read_timeout: reading and writing, select returned %d, errno %d",
			m, err );
		if( m < 0 ){
			/* error */
			if( err != EINTR ){
				Errorcode = JFAIL;
				LOGERR_DIE( _("Write_read_timeout: select error") );
			}
			continue;
		} else if( m == 0 ){
			/* timeout */
			continue;
		}
		if( Status && FD_ISSET( 1, &readfds ) ){
			n = Read_status_timeout(-1);
			DEBUG1("Write_read_timeout: Read_status_timeout returned %d", n );
			if( n == -1 ){
				LOGERR(_("Write_read_timeout: read from printer failed"));
				len = -1;
				break;
			}
		}
		if( FD_ISSET( 1, &writefds ) ){
			DEBUG4("Write_read_timeout: write possible on fd %d", 1 );
			Set_nonblock_io( 1 );
			errno = 0;
			m = write( 1, buffer, len );
			err = errno;
			Set_block_io( 1 );
			DEBUG4("Write_read_timeout: write() returned %d", m );
			errno = err;
			if( m < 0 ){
				if( 0
#if defined(EWOULDBLOCK)
					|| err == EWOULDBLOCK
#endif
#if defined(EAGAIN)
					|| err == EAGAIN
#endif
#if !defined(EWOULDBLOCK) && !defined(AGAIN)
#error  No definition for EWOULDBLOCK or EAGAIN
#endif
					){
					DEBUG4( "Write_read_timeout: write would block but select says not!" );
					errno = err;
				} else if( err == EINTR ){
					/* interrupted, nothing written */
					DEBUG4( "Write_read_timeout: write interrupted" );
					errno = err;
				} else {
					LOGERR( _("Write_read_timeout: write error") );
					len = -1;
					errno = err;
				}
			} else {
				len -= m;
				buffer += m;
				if( len == 0 ){
					break;
				}
			}
		}
	}
	DEBUG4("Write_read_timeout: returning %d", len );
	return( len );
}

/*
 * int Read_status_timeout( int timeout )
 * Read status information from printer
 *
 *  if timeout == -1 then we do a quick poll
 *    we are called with this only if there is quaranteed
 *    data to read OR we will do a nonblocking read OR we will
 *    timeout REALLY quickly... 1 second.  This is ugly ugly ugly
 *    but the only way that you can make polling a parallel port
 *    under Linux work.
 *
 *  if timeout < 0  then we do not block
 *  if timeout == 0 then we block forever
 *  if timeout > 0 then block for timeout
 *
 *  Return:
 *    0   - status read
 *    -1  - error or EOF
 *    -2  - timeout
 */

int Read_status_timeout( int timeout )
{
	int count, err;
	char monitorbuff[SMALLBUFFER];

	monitorbuff[0] = 0;
	/* we read from stdout */
	
	err = 0;
	/* if we have timeout < 0, and are not polling then we
	 * do nonblocking read.
	 */
	if( timeout < 0 ){
		errno = 0;
		Set_nonblock_io( 1 );
		count = read( 1, monitorbuff, sizeof(monitorbuff) - 1 );
		Set_block_io( 1 );
		if( count == -1 && ( 0
#if				 defined(EWOULDBLOCK)
				|| err == EWOULDBLOCK
#endif
#if				 defined(EAGAIN)
				|| err == EAGAIN
#endif
		)) count = 0; 
	} else if( Set_timeout() ){
		Set_timeout_alarm( timeout  );
		count = read( 1, monitorbuff, sizeof(monitorbuff) - 1 );
		err = errno;
	} else {
		count = -2;
		errno = EINTR;
		err = errno;
	}
	Clear_timeout();

	DEBUG1("Read_status_timeout: timeout %d, count %d",
		timeout, count );

	if( count > 0 ){
		monitorbuff[count] = 0;
		DEBUG1("Read_status_timeout: read count %d, '%s'", count, monitorbuff );
		Put_inbuf_len( monitorbuff, count );
		Get_inbuf_str();
		count = 0;
	} else if( count == 0 ){
		if( timeout < 0 ){
			/* handles the polling case */
			DEBUG1("Read_status_timeout: no status read, timeout %d", timeout);
		} else if( Poll_for_status ){
			/* handles the status case with the timeout */
			DEBUG1("Read_status_timeout: no polled status, waiting %d msec", Dev_sleep);
			plp_usleep( Dev_sleep * 1000 );
		} else if( Ignore_eof ){
			DEBUG1("Read_status_timeout: no status read, ignoring EOF");
		} else {
			DEBUG1("Read_status_timeout: no status read, EOF");
			count = -1;
		}
	}
	return( count );
}

/*
 * void Resolve_key_val( char *prefix, char *key_val,
 *	Wr_out routine )
 * - determine what to do for the specified key and value
 *   we first check to see if it is a builtin
 *   if it is, then we do not need to process further
 */

 struct line_list setvals;

void Resolve_key_val( char *prefix, char *key_val, Wr_out routine )
{
	char *value, *id = 0, *prefix_id = 0, *s;
	int c = 0, done = 0;
	int depth = setvals.count;

	/* we expand the key here */
	key_val = Fix_option_str( key_val, 0, 0, 0);
	if( (value = safestrpbrk( key_val, Value_sep )) ){
		c = *value; *value = 0;
	}
	id = safestrdup( key_val, __FILE__,__LINE__ );
	lowercase( id );
	if( value ){
		*value++ = c;
		if( c == '=' ){
			Check_max(&setvals,1);
			setvals.list[setvals.count++] = key_val;
		}
	}
	/* decide if it is a built-in function or simply forcing a string out */
	DEBUG4("Resolve_key_val: prefix '%s', id '%s'='%s'",
		prefix,id,value);
	if( !done ){
		done = Builtin( prefix, id, value, routine);
		DEBUG4("Resolve_key_val: '%s' is builtin '%d'", id, done );
	}
	if( !done && !strncasecmp(prefix,"pjl", 3) ){
		done = Pjl_setvar( prefix, id, value, routine);
		DEBUG4("Resolve_key_val: '%s' is PJL '%d'", id, done );
	}
	if( !done ){
		prefix_id = safestrdup2( prefix, id,__FILE__,__LINE__);
		if( (s = Find_str_value( &Model, prefix_id, Value_sep )) ){
			DEBUG4("Resolve_key_val: string value of id '%s'='%s'", id, s);
			if( *s == '[' ){
				struct line_list l;

				Init_line_list( &l );
				Split( &l, s, List_sep, 0, 0, 0, 1, 0 );
				Resolve_list( prefix, &l, routine );
				Free_line_list( &l );
				done = 1;
			} else {
				routine( s );
				done = 1;
			}
		}
		if( prefix_id ) free(prefix_id); prefix_id = 0;
	}
	DEBUG4("Resolve_key_val: done '%d', value '%s'", done, value);
	if( !done && value && *value && !Is_flag(value, 0) ){
		DEBUG4("Resolve_key_val: after Is_flag value now '%s'", value);
		routine( Fix_val(value) );
		done = 1;
	}
	setvals.count = depth;
	if(id)free(id); id = 0;
	if(key_val)free(key_val); key_val = 0;
}

int Is_flag( char *s, int *v )
{
	int n = 0;
	if( s && *s && strlen(s) == 1 && isdigit(cval(s)) &&
		(n = (*s == '0' || *s == '1' || *s == '@')) && v ){
		if( *s == '@' ){
			*v = 0;
		} else {
			*v = *s - '0';
		}
	}
	return( n );
}

/*
 * void Resolve_list( char *prefix, char **list, int count, W_out routine )
 * Resolve the actions specified by a list
 * List has the form [ xx xx xx or [xx xx xx
 */

void Resolve_list( char *prefix, struct line_list *l, Wr_out routine )
{
	int i;
	char *s;
	DEBUG4("Resolve_list: prefix '%s', count %d", prefix, l->count );
	if(DEBUGL4)Dump_line_list("Resolve_list",l);
	for( i = 0; i < l->count; ++i ){
		s = l->list[i];
		DEBUG4("Resolve_list: prefix '%s', [%d]='%s'", prefix, i, s );
		if( (*s == '[') || (*s == ']')) ++s;
		if( *s ){
			Resolve_key_val( prefix, s, routine );
		}
	}
}

/*
 * void Resolve_user_opts( char *prefix, struct line_list *only, *opts,
 *		W_out routine )
 * Resolve the actions specified by a the user and allowed as options
 * 
 */

void Resolve_user_opts( char *prefix, struct line_list *only,
	struct line_list *l, Wr_out routine )
{
	int i, c = 0, cmp;
	char *id = 0, *s, *value;
	DEBUG4("Resolve_user_opts: prefix '%s'", prefix );
	if(DEBUGL4)Dump_line_list("Resolve_user_opts - only",only);
	if(DEBUGL4)Dump_line_list("Resolve_user_opts - list",l);
	for( i = 0; i < l->count; ++i ){
		if( id ) free(id); id = 0;
		s = l->list[i];
		if( s == 0 || *s == 0 ) continue;
		id = safestrdup(s,__FILE__,__LINE__);
		if( (value = safestrpbrk( id, Value_sep )) ){ c = *value; *value = 0; }
		cmp = Find_first_key( only, id, Value_sep, 0 );
		DEBUG4("Resolve_user_opts: id '%s', Find_first=%d, value '%s'",
			id, cmp, value?(value+1):0 );
		if( value ){ *value = c; }
		if( cmp == 0 ){
			Resolve_key_val( prefix, id, routine );
		}
	}
	if( id ) free(id); id = 0;
}


/*
 * char *Fix_option_str( char *str, int remove_ws, int trim, int one_line )
 *  if( remove_ws ) then remove all isspace() characters
 *  if( trim) find and remove all white space before and after \n
 *  if( one_line) change \n to space
 *  do escape substitutions
 */

char *Fix_option_str( char *str, int remove_ws, int trim, int one_line )
{
	char *s, *t, *value, *tempvalue, *format_char;
	char num[LARGEBUFFER];
	char fmt[127];
	int c, insert, len, lookup;
	char *sq, *eq;

	if( ISNULL(str) ){ return 0; }
	str = safestrdup( str,__FILE__,__LINE__);
	DEBUG5("Fix_option_str: orig '%s', remove_ws %d, trim %d, one_line %d",
		str, remove_ws, trim, one_line );
	tempvalue = 0;
	/* now we do escape sequences */
	for( s = str; (s = safestrchr( s, '\\' )); ){
		/* \xYY -> xYY
           0123    0123
		*/
           
		c = s[1];
		len = 1;
		insert = 0;

		DEBUG4("Fix_option_str: escape at '%s', '%c'", s, c );
		s[0] = c;
		if( isalpha(c) ){
			switch( c ){
			case 'r': s[0] = '\r'; break;
			case 'n': s[0] = '\n'; break;
			case 't': s[0] = '\t'; break;
			case 'f': s[0] = '\f'; break;
			}
			goto done;
		} else if( isdigit( c ) ){
			for( len = 0;
				len < 3 && (s[len] = s[1+len]) && isdigit(cval(s+len));
				++len );
			s[len] = 0;
			DEBUG5("Fix_option_str: octal str '%s', end '%s'", s, s+len+1 );

			s[0] = strtol( s, 0, 8 );
			goto done;
		} else if( c != '%' ){
  done:
			memmove( s+1, s+len+1, strlen(s+len+1)+1);
			++s;
			DEBUG5("Fix_option_str: after move '%s'", s );
		} else {
			/* \%FMT{...} or \%FMT[]
				s[0] == \
				s[1] == %
				s[2] == FMT
				s[2+strlen(FMT)] is bracket
			*/
			DEBUG5("Fix_option_str: var sub into '%s'", s+1 );
			if( (sq = safestrpbrk( s, "[{" )) ){
				if(tempvalue) free(tempvalue); tempvalue = 0;
				lookup = c = sq[0];
				*sq++ = 0;
				eq = safestrpbrk( sq, "]}" );
				if( eq ){
					*eq++ = 0;
				} else {
					eq = sq + strlen(sq);
				}
				if( safestrchr(s+2, '%') ){
					Errorcode = JABORT;
					FATAL(_("Fix_option_str: bad form '%s'"), s );
				}
				SNPRINTF(fmt, sizeof(fmt))  "%%%s", s+2 );
				DEBUG5("Fix_option_str: getting '%s' fmt '%s'", sq, fmt );
				format_char = t = fmt+strlen(fmt)-1; /* this is safe */
				c = cval(t);
				if( !safestrcmp( sq, "ARGV") ){
					tempvalue = value = Join_vector( Argv+1, Argc, " ");
				} else {
					/* find the value of the variable */
					value = Find_sub_value( lookup, sq, c == 's' );
					if( value == 0 || *value == 0 ){
						value = "";
					} else {
						tempvalue = value = safestrdup(value,__FILE__,__LINE__);
						DEBUG3(
							"Fix_option_str: using fmt '%s', type '%c', value '%s'",
							fmt, c, value );
						/* trim the string */
						while(isspace(cval(value))) ++value;
						for( t = value+strlen(value); t > value && isspace(cval(t-1)); --t );
						*t = 0;
						if( value[0] == '[' ){
							++value;
							while(isspace(cval(value))) ++value;
							if( t > value && cval(t-1) == ']' ) *t = 0;
							for( t = value+strlen(value); t > value && isspace(cval(t-1)); --t );
						}
					}
					if( *value ){
						DEBUG3("Fix_option_str: fixing value '%s'", value );
						t = Fix_option_str( value, remove_ws, trim, one_line );
						if( tempvalue ) free(tempvalue);
						value = tempvalue = t;
					}
				}
				DEBUG3("Fix_option_str: fixed value '%s'", value );
				num[0] = 0;
				switch( c ){
				case 'o': case 'd': case 'x':
					SNPRINTF(num, sizeof(num))  fmt, strtol(value, 0, 0));
					break;
				case 'f': case 'g': case 'e':
					SNPRINTF(num, sizeof(num))  fmt, strtod(value, 0));
					break;
				case 's':
					SNPRINTF(num, sizeof(num))  fmt, value);
					break;
				case 'U':
					*format_char = 's';
					SNPRINTF(num, sizeof(num))  fmt, value);
					uppercase(num);
					break;
				case 'M':
					*format_char = 's';
					SNPRINTF(num, sizeof(num))  fmt, value);
					lowercase(num);
					if( num[0] && islower(cval(num)) ) num[0] = toupper(cval(num));
					break;
				case 'L':
					*format_char = 's';
					SNPRINTF(num, sizeof(num))  fmt, value);
					lowercase(num);
					break;
				}
				if( tempvalue ) free(tempvalue); tempvalue = 0;
				*s = 0;
				len = strlen(str)+strlen(num);
				DEBUG5("Fix_option_str: first '%s', num '%s', rest '%s'",
					str, num, eq );
				s = str;
				str = safestrdup3(str,num,eq,__FILE__,__LINE__);
				free(s);
				s = str+len;
				DEBUG5("Fix_option_str: result '%s', next '%s'", str, s );
			}
		}
	}
	if( remove_ws ){
		for( s = t = str; (*t = *s); ++s ){
			if( !isspace(cval(t)) ) ++t;
		}
		*t = 0;
	}
	if( trim ){
		while( isspace(cval(str)) ) memmove( str, str+1, strlen(str+1)+1 );
		for( s = str; (s = safestrpbrk( s, "\n" )); s = t ){
			for( t = s; t > str && isspace(cval(t-1)); --t );
			if( t != s ){
				memmove( t, s, strlen(s)+1 );
				s = t;
			}
			for( t = s+1; isspace(cval(t)); ++t );
			if( t != s+1 ){
				memmove( s+1, t, strlen(t)+1 );
				t = s+1;
			}
		}
	}
	if( one_line ){
		for( s = str; (s = safestrpbrk( s, "\t\n" )); ++s ){
			*s = ' ';
		}
	}
	if( ISNULL(str) ){
		if( str ) free(str); str = 0;
	}
	DEBUG4("Fix_option_str: returning '%s'", str );
	return( str );
}

/*
 * char *Find_sub_value( int c, char *id, char *sep )
 *  if we have ?xxx?v1:v2 then if xxx has a value or
 *    the value is not '0' then we look up v1 else
 *    we look up v2
 *  if c == {, try the user -Z opts
 *  try user -T opts
 *  if id is single character try options
 *  try Model config
 *  default value is 0
 */

char *Find_sub_value( int c, char *id, int strval )
{
	char *s = 0;
	char *v1, *v2;
	char copy[128];
	int i;
	DEBUG4("Find_sub_value: type '%c', id '%s'", c, id );
	if( *id == '?' ){
		strncpy(copy,id+1,sizeof(copy));
		v1 = v2 = 0;
		v1 = safestrchr(copy, '?');
		if( v1 ){
			v2 = safestrchr(v1+1, ':');
		}
		if( v2 == 0 ){
			FATAL(_("Find_sub_value: bad conditional format '%s'"), id );
		}
		*v1++ = 0;
		*v2++ = 0;
		s = Find_sub_value( c, copy, 1 );
		DEBUG4("Find_sub_value: CONDITIONAL id '%s'='%s'", copy, s );
		if( !ISNULL(s) ){
			s = Find_sub_value( c, v1, 1 );
		} else {
			s = Find_sub_value( c, v2, 1 );
		}
		DEBUG4("Find_sub_value: CONDITIONAL type '%c', id '%s'='%s'", c, id, s );
		return( s );
	}
	if( s == 0 && !strcasecmp(id,"pid") ){
		static char pid[32];
		SNPRINTF(pid, sizeof(pid)) "%d",getpid());
		s = pid;
	}
	if( s == 0 && !strcasecmp(id, "papersize" ) ){
		struct line_list l;
		Init_line_list(&l);
		s = Find_exists_value( &Model, "papersize", Value_sep );
		DEBUG4("Find_sub_value: papersize '%s'", s );
		Split( &l, s, List_sep, 1, Value_sep, 1, 1, 1 );
		/* we first check the Zopts - in reverse order */
		v1 = v2 = 0;
		for( i = Unsorted_Zopts.count-1; i >= 0 ; --i ){
			v1 = Unsorted_Zopts.list[i];
			if( (v2 = safestrstr(v1,"paper="))
				|| (v2 = safestrstr(v1,"papersize=")) ){
				v1 = safestrchr(v1,'=');
				++v1;
				break;
			} else {
				v2 = Find_exists_value(&l,v1,Value_sep);
			}
			DEBUG4("Find_sub_value: Zopt '%s', found '%s'", v1, v2 );
			if( v2 ){
				break;
			}
		}
		if( v2 == 0 ) for( i = Unsorted_Topts.count-1; i >= 0 ; --i ){
			v1 = Unsorted_Topts.list[i];
			if( (v2 = safestrstr(v1,"paper="))
				|| (v2 = safestrstr(v1,"papersize=")) ){
				v1 = safestrchr(v1,'=');
				++v1;
				break;
			} else {
				v2 = Find_exists_value(&l,v1,Value_sep);
			}
			DEBUG4("Find_sub_value: Topt '%s', found '%s'", v1, v2 );
			if( v2 ){
				break;
			}
		}
		DEBUG4("Find_sub_value: found size '%s' - '%s'", v1, v2 );
		if( v2 == 0 ){
			s = Find_exists_value( &Model, "papersize_default", Value_sep );
		} else {
			s = v1;
		}
		Free_line_list(&l);
	}
	if( s == 0 && setvals.count ){
		char *t, *u;
		int n;
		n = strlen(id);
		for( i = 0; i < setvals.count; ++i ){
			if( !strncasecmp( (t = setvals.list[i]), id, n) ){
				u = safestrpbrk(t,Value_sep);
				if( (u - t) == n ){
					s = u+1;
					DEBUG4("Find_sub_value: from setvals '%s', value '%s'",
						u,s);
					break;
				}
			}
		}
	}
	if( s == 0 && c == '{' ){
		s = Find_exists_value( &Zopts, id, Value_sep );
		DEBUG4("Find_sub_value: from Zopts '%s'", s );
	}
	if( s == 0 ){
		s = Find_exists_value( &Model, id, Value_sep );
		DEBUG4("Find_sub_value: from Model '%s'", s );
	}
	if( s == 0 ){
		if( strval ){
			s = "";
		} else {
			s = "0";
		}
	}
	DEBUG4("Find_sub_value: type '%c', id '%s'='%s'", c, id, s );
	return( s );
}

/*
 * int Builtin( char* prefix, char *pr_id, char *value, Wr_out routine)
 *  - check to see if the pr_id is for a builtin operation
 *  - call the built-in with the parameters
 *  - if found, the invoked return should return return 1 
 *        so no further processing is done
 *    if not found or the return value is 0, then further processing is
 *    done.
 */

int Builtin( char* prefix, char *id, char *value, Wr_out routine)
{
	int cmp=1;
	char *s, *prefix_id = 0;
	Builtin_func r;
	struct keyvalue *v;

	DEBUG4("Builtin: looking for '%s'", id );
	for( v = Builtin_values;
		(s = v->key) && (cmp = strcasecmp(s, id));
		++v );
	if( cmp && prefix && *prefix ){
		prefix_id = safestrdup2( prefix, id,__FILE__,__LINE__);
		DEBUG4("Builtin: looking for '%s'", prefix_id );
		for( v = Builtin_values;
			(s = v->key) && (cmp = strcasecmp(s, prefix_id));
			++v );
	}
	if( cmp == 0 ){
		DEBUG4("Builtin: found '%s'", s );
		r = (Builtin_func)(v->var);
		cmp = (r)( prefix, id, value, routine);
	} else {
		cmp = 0;
	}
	DEBUG4("Builtin: returning '%d'", cmp );
	if( prefix_id){ free(prefix_id); prefix_id = 0; }
	return( cmp );
}

/***************************************************************************
 * Decode_status (plp_status_t *status)
 * returns a printable string encoding return status
 ***************************************************************************/

const char *Decode_status (plp_status_t *status)
{
    static char msg[128];

	int n;
    *msg = 0;		/* just in case! */
    if (WIFEXITED (*status)) {
		n = WEXITSTATUS(*status);
		if( n > 0 && n < 32 ) n += JFAIL-1;
		(void) SNPRINTF (msg, sizeof(msg))
		"exit status %d", WEXITSTATUS(*status) );
    } else if (WIFSTOPPED (*status)) {
		(void) strcpy(msg, "stopped");
    } else {
		(void) SNPRINTF (msg, sizeof(msg)) "died%s", WCOREDUMP (*status) ? " and dumped core" : "");
		if (WTERMSIG (*status)) {
			(void) SNPRINTF(msg+strlen(msg), sizeof(msg)-strlen(msg))
				 ", %s", Sigstr ((int) WTERMSIG (*status)));
		}
    }
    return (msg);
}

/***************************************************************************
 * plp_usleep() with select - simple minded way to avoid problems
 ***************************************************************************/
int plp_usleep( int i )
{
	struct timeval t;
	DEBUG3("plp_usleep: starting usleep %d", i );
	if( i > 0 ){
		memset( &t, 0, sizeof(t) );
		t.tv_usec = i%1000000;
		t.tv_sec = i/1000000;
		DEBUG3("plp_usleep: %d sec, %d microsec", t.tv_sec, t.tv_usec );
		i = select( 0,
			FD_SET_FIX((fd_set *))(0),
			FD_SET_FIX((fd_set *))(0),
			FD_SET_FIX((fd_set *))(0),
			&t );
		DEBUG3("plp_usleep: select done, status %d, errno %d, '%s'", i, errno, Errormsg(errno) );
	}
	return( i );
}

void Strip_leading_spaces ( char **vp )
{
	char *s = *vp;
	if( s ){
		do{
			while( isspace( cval(s) ) ) memmove(s,s+1,strlen(s+1)+1); 
			if( (s = safestrchr( s, '\n' )) ) ++s;
		} while(s);
		if( **vp == 0 ) *vp = 0;
	}
}
  

int Font_download( char* prefix, char *id, char *value, Wr_out routine)
{
	char *dirname = 0, *filename, *s;
	char buffer[LARGEBUFFER];
	int n, fd, i;
	struct line_list fontnames;

	DEBUG2("Font_download: prefix '%s', id '%s', value '%s'",prefix, id, value );

	Init_line_list(&fontnames);
	s = safestrdup2(prefix,"fontdir",__FILE__,__LINE__);
	DEBUG4("Font_download: directory name '%s'", s );
	dirname = Find_str_value( &Model, s, Value_sep );
	if(s) free(s); s = 0;
	if( !dirname ){
		LOGERR(_("Font_download: no value for %sfontdir"), prefix);
		return(1);
	}

	/* if there is a value, then use it */
	if( (filename = Fix_option_str(value,1,1,1)) ){
		DEBUG2("Font_download: fixed value '%s'", filename );
		Split(&fontnames,filename,Filesep,0,0,0,0,0);
		if( filename ) free(filename); filename = 0;
	} else {
		/* if there is a 'font' value, then use it */
		filename = Find_sub_value( '{', "font", 1 );
		Split(&fontnames,filename,Filesep,0,0,0,0,0);
	}
	if( fontnames.count == 0 ){
		LOGMSG(_("Font_download: no 'font' value"));
		return(1);
	}
	filename = 0;
	for( i = 0; i < fontnames.count; ++i ){
		if( filename ) free(filename); filename = 0;
		if( (s = safestrrchr(fontnames.list[i],'/')) ){
			++s;
		} else {
			s = fontnames.list[i];
		}
		filename = safestrdup3(dirname,"/",s,__FILE__,__LINE__);
		/* get rid of // */
		DEBUG4("Font_download: filename '%s'", filename );
		if( (fd = open( filename, O_RDONLY )) >= 0 ){
			Set_max_fd(fd);
			while( (n = read(fd, buffer,sizeof(buffer))) >0 ){
				Put_outbuf_len( buffer, n );
			}
			close(fd);
		} else {
			LOGERR("Font_download: cannot open '%s'", filename);
		}
	}
	if( filename ) free(filename); filename = 0;
	Free_line_list(&fontnames);
	return(1);
}


/* by hlauer: Hermann Lauer
 * Bildverarbeitungsgruppe des Interdiziplinaeren Zentrums fuer
 * wissenschaftliches Rechnen, Universitaet Heidelberg
 * INF 368; 69120 Heidelberg; Tel: (06221)54-8826, -6314  Fax: -8850
 * Email: Hermann.Lauer@iwr.uni-heidelberg.de
 * Use a common subroutine to build DISPLAY part - the job start
 *  could use this too.
 *  if console@, returns an empty string
 */

 char *Jobstart_str="@PJL JOB NAME = \"%s\"";
 char *Job_display=" DISPLAY = \"%s\" ";
 char *Jobend_str="@PJL EOJ NAME = \"%s\"";
 char *Jobreset_str="@PJL RESET\n";

 static void Pjl_displayname(void)
{
	char msg[64];
	char *s = 0;
	msg[0] = 0;
	if( Pjl_ready_msg ){
		DEBUG2("Pjl_displayname: pjl_ready_msg '%s'", Pjl_ready_msg);
		if( (s = Fix_option_str( Pjl_ready_msg, 0, 1, 1)) ){
			SNPRINTF(msg,sizeof(msg))  "%s", s);
			free(s); s = 0;
			if( msg[0] ) s = msg;
		}
	}
	if( !s ) s = Loweropts['n'-'a'];  /* try to get user or jobname */
	if( !s ) s = Upperopts['J'-'A'];
	if( !s ){
		SNPRINTF(msg,sizeof(msg))  "PID %d", getpid());
		s = msg;
	}
	SNPRINTF(Jobname,sizeof(Jobname))  "%s", s);
	DEBUG2("Pjl_displayname: Jobname '%s'", Jobname);
	msg[0] = 0;
	if( Pjl_done_msg ){
		DEBUG2("Pjl_displayname: pjl_done_msg '%s'", Pjl_done_msg);
		if( (s = Fix_option_str( Pjl_done_msg, 0, 1, 1)) ){
			SNPRINTF(msg,sizeof(msg))  "%s", s);
		}
	}
	SNPRINTF(End_ready_msg,sizeof(End_ready_msg))  "%s", msg);
	DEBUG2("Pjl_displayname: End_ready_msg '%s'", End_ready_msg);
}

/*
 * int Pjl_job()
 *  - put out the JOB START string
 */

void Pjl_job()
{
	char *s, buffer[SMALLBUFFER];
	int n = 0, len;
	buffer[0] = 0;

	if( (s = Find_value(&Model,"pjl_job",Value_sep )) ){
		Is_flag(s, &n);
	}
	if( n && ( Find_first_key( &Pjl_only, "job", 0, 0)
			|| !Find_first_key( &Pjl_except, "job", 0, 0)) ){
		n = 0;
	}
	DEBUG2("Pjl_job: Pjl %d, job '%s', flag %d", Pjl, s, n );
	if( Pjl == 0 || n == 0 ){
		return;
	}
	buffer[0] = 0;
	len = strlen(buffer);
	SNPRINTF(buffer+len, sizeof(buffer)-len) Jobreset_str);
	len = strlen(buffer);
	SNPRINTF(buffer+len, sizeof(buffer)-len) Jobstart_str, Jobname );
	len = strlen(buffer);
	SNPRINTF(buffer+len, sizeof(buffer)-len) Job_display, Jobname );

	if( (s = Find_exists_value( &Zopts, "startpage", Value_sep))
		|| (s = Find_exists_value( &Topts, "startpage", Value_sep)) ){
		n = atoi( s );
		if( n > 0 ){
			len = strlen(buffer);
			SNPRINTF(buffer+len, sizeof(buffer)-len)  " START = \"%d\"", n );
		}
	}
	if( (s = Find_exists_value( &Zopts, "endpage", Value_sep))
		|| (s = Find_exists_value( &Topts, "endpage", Value_sep)) ){
		n = atoi( s );
		if( n > 0 ){
			len = strlen(buffer); 
			SNPRINTF(buffer+len, sizeof(buffer)-len)  " END = \"%d\"", n );
		}
	}
	DEBUG2("Pjl_job: final = '%s'", buffer );
	Put_pjl( buffer );
}

/*
 * int Pjl_eoj(char *name)
 *  - put out the JOB EOJ string
 */

void Pjl_eoj(char *name)
{
	char buffer[SMALLBUFFER];
	char *s;
	int n = 0;
	int len;
	if( (s = Find_value(&Model,"pjl_job",Value_sep )) ){
		Is_flag(s, &n);
	}
	if( n && ( Find_first_key( &Pjl_only, "job", 0, 0)
			|| !Find_first_key( &Pjl_except, "job", 0, 0)) ){
		n = 0;
	}
	DEBUG2("Pjl_eoj: Pjl %d, job '%s', flag %d", Pjl, s, n );
	if( Pjl == 0 || n == 0 ){
		return;
	}
	buffer[0] = 0;
	len = strlen(buffer);
	SNPRINTF(buffer+len, sizeof(buffer)-len) Jobreset_str);
	len = strlen(buffer);
	SNPRINTF(buffer+len, sizeof(buffer)-len) Jobend_str, name );
	Put_pjl( buffer );
}

/*
 * PJL RDYMSG option
 *  console    enables messages on console
 *  console@   disables or erases messages on console
 */

 char *PJL_RDYMSG_str  = "@PJL RDYMSG DISPLAY = \"%s\" ";

void Pjl_console_msg( int start )
{
	char buffer[SMALLBUFFER], name[SMALLBUFFER];

	DEBUG2("Pjl_console: flag %d, start %d, jobname '%s ",
		Pjl_console, start, Jobname );
	if( Pjl == 0 || Pjl_console == 0 ){
		return;
	}
	name[0] = 0;
	Pjl_displayname();
	if( start ){
		SNPRINTF(name, sizeof(name))  "%s", Jobname );
	} else {
		SNPRINTF(name, sizeof(name))  "%s", End_ready_msg );
	}
	if( Pjl_display_size > 0 &&  strlen(name) > Pjl_display_size ){
		name[Pjl_display_size] = 0;
	}
	SNPRINTF(buffer, sizeof(buffer))  PJL_RDYMSG_str, name );
	DEBUG2("Pjl_console: console msg '%s'", buffer);
	Put_pjl( buffer );
}

/*
 * PJL SET VAR=VALUE
 *  1. The variable must be in the Pjl_vars_set list
 *     This list has entries of the form:
 *        var=value var=value
 *        var@  -> var=OFF  
 *        var   -> var=ON  
 *  2. The variable must not be in the Pjl_vars_except list
 *     This list has entries of the form
 *        var  var var
 */

 char *PJL_SET_str = "@PJL SET %s = %s";

int Pjl_setvar(char *prefix, char*id, char *value, Wr_out routine)
{
	int found = 0, i;
	char buffer[SMALLBUFFER], *s;

	DEBUG3( "Pjl_setvar: prefix '%s', id '%s', value '%s'",
		prefix, id, value );
	if( !Find_first_key( &Pjl_vars_set, id, Value_sep, &i ) ){
		s = safestrpbrk( Pjl_vars_set.list[i], Value_sep );
		DEBUG3( "Pjl_setvar: found id '%s'='%s'", id, s );
		if( Find_first_key( &Pjl_vars_except, id, Value_sep, 0 ) ){
			DEBUG3( "Pjl_setvar: not in the except list" );
			if( value == 0 || *value == 0 ){
				value = s;
			}
			if( value == 0 || *value == 0 ){
				value = "ON";
			} else if( *value =='@' ){
				value = "OFF";
			} else if( *value == '=' || *value == '#' ){
				/* skip over = and # */
				++value;
			}
			DEBUG3( "Pjl_setvar: setting '%s' = '%s'", id, value );
			SNPRINTF(buffer,sizeof(buffer)) PJL_SET_str,id,value);
			uppercase(buffer);
			routine(buffer);
			found = 1;
		}
	}
	return(found);
}

int Pcl_setvar(char *prefix, char*id, char *value, Wr_out routine )
{
	int found = 0, i;
	char *s;

	DEBUG3( "Pcl_setvar: prefix '%s', id '%s', value '%s'",
		prefix, id, value );
	if( !Find_first_key( &Pcl_vars_set, id, Value_sep, &i ) ){
		s = safestrpbrk( Pcl_vars_set.list[i], Value_sep );
		DEBUG3( "Pcl_setvar: found id '%s'='%s'", id, s );
		if( Find_first_key( &Pcl_vars_except, id, Value_sep, 0 ) ){
			DEBUG3( "Pcl_setvar: not in the except list" );
			value = s;
			if( value == 0 || *value == 0 ){
				value = 0;
			} else if( *value =='@' ){
				value = 0;
			} else if( *value == '=' || *value == '#' ){
				/* skip over = and # */
				++value;
			}
			DEBUG3( "Pcl_setvar: setting '%s' = '%s'", id, value );
			if( value ){
				routine(value);
			}
			found = 1;
		}
	}
	return(found);
}

/*
 * Do_sync()
 *  actual work is done here later
 *  sync=pjl  enables pjl
 *  sync=ps   enables postscript
 *  sync@     disables sync
 *  default is to use the prefix value
 */

 char *PJL_ECHO_str = "@PJL ECHO %s";
 char *CTRL_T = "\024";
 char *CTRL_D = "\004";

/*
 * void Do_sync( int sync_timeout, int sync_interval, int pagecount_pjl )
 * - get a response from the printer
 *  timeout is sync_timeout
 *  send the synchronization stuff at sync_interval seconds
 *  pagecount_pjl means that we should check for pagecount
 */

void Do_sync( int sync_timeout, int sync_interval, int pagecount_pjl )
{
	char buffer[SMALLBUFFER], name[SMALLBUFFER], *s, *sync_str;
	int len, elapsed, timeout, sync, use, use_ps, use_pjl,
		cx, attempt;
	time_t start_t, current_t, interval_t;
	struct stat statb;

	time( &start_t );

	cx = 0;
	sync_str = 0;

	s = Sync;
	DEBUG2("Do_sync: sync is '%s'", Sync );

	/* we see if we can do sync */
	use = use_pjl = use_ps = 0;
	if( !Is_flag( s, &use ) ){
		/* we check to see if the string is specifying */
		if( !strcasecmp(s,"pjl") ){
			use_pjl = Pjl;
		}
		if( !strcasecmp(s,"ps") ){
			use_ps = Ps;
		}
		use = 1;
	} else if( use ){
		use_ps = Ps;
		use_pjl = Pjl;
	} else {
		return;
	}
	if( use_pjl
		&& ( Find_first_key( &Pjl_only, "echo", 0, 0)
			|| !Find_first_key( &Pjl_except, "echo", 0, 0)) ){
		use_pjl = 0;
	}
	if( use_pjl ) use_ps = 0;

	if( !use_ps && !use_pjl ){
		Errorcode = JABORT;
		FATAL(_("Do_sync: sync '%s' and method not supported"), s );
	}

	if( !sync_str && use_pjl ) sync_str = "pjl echo";
	if( !sync_str && use_ps ) sync_str = "ps";

	LOGMSG(_("Do_sync: getting sync using '%s'"), sync_str );

	attempt = 0;
	sync = 0;

	if( Appsocket && fstat(1,&statb) == -1 ){
		Open_device( Device );
		DEBUG1("Start_of_job: Appsocket device open done");
	}

 again:

	++attempt;
	DEBUG2("Do_sync: attempt %d", attempt );

	/* get start time */
	time( &interval_t );
	Init_outbuf();
	SNPRINTF(name, sizeof(name))  "%d@%s", getpid(), Time_str(0,0));
	for( s = name; ( s = safestrchr(s,':')); ++s) *s = '-';
	DEBUG2("Do_sync: Ps_status_code '%s', use_pjl %d, use_ps %d",
			Ps_status_code, use_pjl, use_ps );
	if( use_pjl ){
		Put_outbuf_str( PJL_UEL_str );
		Put_outbuf_str( PJL_str );
		if( (s = Find_str_value( &Model, "pjl_init", Value_sep)) ){
			struct line_list l;
			Init_line_list(&l);
			DEBUG1("Start_of_job: 'pjl_init'='%s'", s);
			Split( &l, s, List_sep, 0, 0, 0, 1, 0 );
			Resolve_list( "pjl_", &l, Put_pjl );
			Free_line_list( &l );
		}
		SNPRINTF(buffer, sizeof(buffer)) 
			PJL_ECHO_str, name );
		Put_pjl( buffer );

		SNPRINTF(buffer, sizeof(buffer)) PJL_RDYMSG_str, Jobname ); Put_pjl( buffer );

		if( pagecount_pjl ){
			Put_pjl( PJL_INFO_PAGECOUNT_str );
		}
		Put_outbuf_str( PJL_UEL_str );
	} else if( use_ps ){
		if( Ps_status_code == 0 ){
			FATAL(_("Do_sync: sync '%s' and no ps_status_code value"), sync_str );
		}
		if(Pjl){
			Put_outbuf_str( PJL_UEL_str );
			if( Pjl_enter ){
				Put_outbuf_str( PJL_str );
				Put_outbuf_str( "@PJL ENTER LANGUAGE = POSTSCRIPT\n" );
			}
		}
		/* watch out for ^D as part of sync */
		if( !Appsocket && Ps_eoj && Ps_eoj_at_start ) Put_outbuf_str( CTRL_D );
		if( (s = safestrstr(Ps_status_code,"NAME")) ){ cx = *s; *s = 0; }
		Put_outbuf_str( Ps_status_code );
		Put_outbuf_str( name );
		if( s ){ *s = cx; s += 4; Put_outbuf_str( s ); };
		Put_outbuf_str("\n");
		/* watch out for ^D as part of sync */
		if( !Appsocket && Ps_eoj && Ps_eoj_at_end ) Put_outbuf_str( CTRL_D );
		if(Pjl){
			Put_outbuf_str( PJL_UEL_str );
		}
	} else {
		Errorcode = JABORT;
		FATAL(_("Do_sync: no way to synchronize printer, need PJL, or PS") );
	}

	DEBUG2("Do_sync: using sync '%s'", Outbuf );

	while( !sync ){
		/* write the string */
		/* check to see if we need to read */
		time( &current_t );
		elapsed = current_t - start_t;

		/* check for master timeout */
		if( sync_timeout > 0 ){
			if( elapsed >= sync_timeout ){
				break;
			}
			timeout = sync_timeout - elapsed;
		} else {
			timeout = 0;
		}
		elapsed = current_t - interval_t;
		DEBUG3("Do_sync: timeout %d, sync_interval %d, elpased %d",
			timeout, sync_interval, elapsed );
		if( sync_interval > 0 ){
			timeout = sync_interval - elapsed;
			if( timeout <= 0 ){
				LOGMSG(_("Do_sync: printer not detected, trying again"));
				LOGMSG(_("Do_sync: Perhaps printer offline or not detected by Operating System"));
				LOGMSG(_("Do_sync: If on a parallel port then parallel port may not be bidirectional"));
				LOGMSG(_("Do_sync: Use the 'status@' filter option to suppress this check"));
				LOGMSG(_("Do_sync: See the ifhp man page or the ifhp HOWTO for details"));
				goto again;
			}
		}
		DEBUG3("Do_sync: waiting for sync, timeout %d", timeout );
		len = 0;
		if( Outlen ){
			len = Write_read_timeout( Outlen, Outbuf, timeout );
			Init_outbuf();
		} else {
			len = Read_status_timeout( timeout );
		}
		DEBUG3("Do_sync: len '%d'", len );
		if( (s = Find_str_value( &Devstatus, "echo", Value_sep))
			&& safestrstr( s, name ) ){
			sync = 1;
		}
		if( len == -1 ){
			break;
		}
	}
	DEBUG2("Do_sync: sync %d", sync );
	if( sync == 0 ){
		Errorcode = JFAIL;
		FATAL("Do_sync: no sync response from printer" );
	}
	LOGMSG(_("Do_sync: sync done"));
}

/*
 * void Do_waitend( int sync_timeout, int interval )
 * - get the job status from the printer
 *   and wait for the end of job
 */

 char *PJL_USTATUS_JOB_str = "@PJL USTATUS JOB = ON";

void Do_waitend( int waitend_timeout, int waitend_interval,
	int waitend_ctrl_t_interval, int banner )
{
	char *sync_str, *s, *t, *u, buffer[SMALLBUFFER], endname[SMALLBUFFER];
	int len, elapsed, timeout, waitend,
		use, use_pjl, use_ps, use_job, c,
		echo_received = 0;
	time_t start_t, current_t, interval_t;
	char *status;

	time( &start_t );

	s = (Waitend?Waitend:Sync);
	DEBUG1("Do_waitend: waitend '%s', end_status '%s', waitend timeout %d, interval %d, ctrl_t_interval %d",
		s, End_status, waitend_timeout, waitend_interval, waitend_ctrl_t_interval
		);
	if( Is_flag( s, &use ) && use == 0 ){
		return;
	}

	DEBUG1("Do_waitend: ps_ctrl_t '%d'", Ps_ctrl_t );

	/* we see if we can do sync */
	use = use_pjl = use_ps = use_job = 0;
	sync_str = 0;
	if( !Is_flag( s, &use ) ){
		/* we check to see if the string is specifying */
		if( !strcasecmp(s,"pjl") ){
			use_pjl = Pjl;
		}
		if( !strcasecmp(s,"ps") ){
			use_ps = Ps;
		}
		use = 1;
	} else if( use ){
		use_job = use_pjl = Pjl;
		use_ps = Ps;
	} else {
		return;
	}

	if( use_pjl ){
		if( (s = Find_value(&Model,"pjl_job",Value_sep )) ){
			Is_flag(s, &use_job);
		}
		if( use_job && ( Find_first_key( &Pjl_only, "job", 0, 0)
				|| !Find_first_key( &Pjl_except, "job", 0, 0)) ){
			use_job = 0;
		}
		if( use_job == 0 && ( Find_first_key( &Pjl_only, "echo", 0, 0)
			|| !Find_first_key( &Pjl_except, "echo", 0, 0)
			) ){
			use_pjl = 0;
		}
	}
	if( !use_ps && !use_pjl ){
		Errorcode = JABORT;
		FATAL(_("Do_waitend: waitend '%s' and method not supported"), s );
	}
	if( use_pjl ) use_ps = 0;
	if( !sync_str && use_job ) sync_str = "pjl job/eoj";
	if( !sync_str && use_pjl ) sync_str = "pjl echo";
	if( !sync_str && use_ps ) sync_str = "ps";

	LOGMSG(_("Do_waitend: getting end using '%s'"), sync_str );

	waitend = 0;

	/* find if we need to periodically send the waitend value */
	DEBUG3("Do_waitend: use_pjl %d, use_job %d, use_ps %d, timeout %d, interval %d",
		use_pjl, use_job, use_ps, waitend_timeout, waitend_interval );

	/* Free_line_list( &Devstatus ); */

	SNPRINTF(endname, sizeof(endname))  "%s PID %d",
		Time_str(1,0), getpid() );
	for( s = endname; ( s = safestrchr(s,':')); ++s) *s = '-';

 again:

	Init_outbuf();
	if( use_pjl ){
		Put_outbuf_str( PJL_UEL_str );
		Put_outbuf_str( PJL_str );
		if( use_job ){
			SNPRINTF(buffer, sizeof(buffer))  Jobstart_str, endname );
			if( Pjl_console ){
				char name[64];
				SNPRINTF(name, sizeof(name))  "%s",Jobname );
				if( Pjl_display_size > 0 &&  strlen(name) > Pjl_display_size ){
					name[Pjl_display_size] = 0;
				}
				len = strlen(buffer);
				SNPRINTF(buffer+len, sizeof(buffer)-len) Job_display, name );
			}
			Put_pjl(buffer);
			Put_pjl( PJL_USTATUS_JOB_str );
			Pjl_eoj(endname);
		} else {
			SNPRINTF(buffer, sizeof(buffer))  PJL_ECHO_str, endname );
			Put_pjl(buffer);
		}
		Put_outbuf_str( PJL_UEL_str );
	} else if( use_ps ){
		if( Ps_status_code == 0 ){
			FATAL("Do_waitend: no ps_status_code value" );
		}
		if( !echo_received  || Ps_ctrl_t ){
			if(Pjl){
				Put_outbuf_str( PJL_UEL_str );
				if( Pjl_enter ){
					Put_outbuf_str( PJL_str );
					Put_outbuf_str( "@PJL ENTER LANGUAGE = POSTSCRIPT\n" );
				}
			}
			if( !Appsocket && Ps_eoj && Ps_eoj_at_start ) Put_outbuf_str( CTRL_D );
			if( !echo_received ){
				c = 0;
				if( (s = safestrstr(Ps_status_code,"NAME")) ){ c = *s; *s = 0; }
				Put_outbuf_str( Ps_status_code );
				Put_outbuf_str( endname );
				if( s ){ *s = c; s += 4; Put_outbuf_str( s ); };
			}
			if( Ps_ctrl_t ){
				Put_outbuf_str( CTRL_T );
			}
			if( !Appsocket && Ps_eoj && Ps_eoj_at_end ) Put_outbuf_str( CTRL_D );
			if(Pjl){
				Put_outbuf_str( PJL_UEL_str );
			}
		}
	}

	time( &interval_t );

	DEBUG3("Do_waitend: sending '%s'", Outbuf );
	while( !waitend ){
		/* write the string */
		/* check to see if we need to read */
		time( &current_t );
		elapsed = current_t - start_t;

		/* check for master timeout */
		if( waitend_timeout > 0 ){
			DEBUG3("Do_waitend: waitend timeout %d, elapsed %d",
				waitend_timeout, elapsed );
			if( elapsed >= waitend_timeout ){
				break;
			}
			timeout = waitend_timeout - elapsed;
		} else {
			timeout = 0;
		}
		if( waitend_interval > 0 ){
			elapsed = current_t - interval_t;
			timeout = waitend_interval - elapsed;
			if( timeout <= 0 ) goto again;
		}
		DEBUG2("Do_waitend: Outlen %d '%s'", Outlen, Outlen?Outbuf:"" );
		if( Outlen ){
			len = Write_read_timeout( Outlen, Outbuf, timeout );
			Init_outbuf();
		} else {
			len = Read_status_timeout( timeout );
		}
		DEBUG3("Do_waitend: len %d", len );
		if( len == -1 ){
			/* we have an error */
			break;
		}
		if(DEBUGL3)Dump_line_list("Do_waitend - Devstatus", &Devstatus );
		if( use_pjl ){
			s = Find_str_value( &Devstatus, "job", Value_sep);
			t = Find_str_value( &Devstatus, "name", Value_sep);
			u = Find_str_value( &Devstatus, "result", Value_sep);
			DEBUG2("Do_waitend: job '%s', name '%s', endname '%s', result '%s'", s, t, u, endname );
			if( s && safestrstr(s,"END") && t && safestrstr(t,endname) ){
				waitend = 1;
			}
			/* we have the job cancelled, so we retry */ 
			if( (safestrstr(s, "CANCELED"))
			  || (safestrstr(u, "CANCELED")) ){
				Set_str_value(&Devstatus,"job",0);
				Set_str_value(&Devstatus,"result",0);
				LOGMSG(_("Do_waitend: job canceled"));
				goto again;
			}
			s = Find_str_value( &Devstatus, "echo", Value_sep);
			DEBUG2("Do_waitend: echo '%s'", s );
			if( safestrstr(s,endname) ){
				waitend = 1;
			}
			DEBUG2("Do_waitend: waitend '%d'", waitend );
		} else if( use_ps ){
			if( !echo_received ){
				s = Find_str_value( &Devstatus, "echo", Value_sep);
				echo_received = (safestrstr( s, endname ) != 0 );
				DEBUG4("Do_waitend: echo '%s', endname '%s', echo_received %d",
					s, endname, echo_received );
			}
			if( echo_received ){
				if( ISNULL(End_status) ){
					waitend = 1;
				} else {
					status = Find_str_value( &Devstatus, "status", Value_sep);
					/* The status we get back must be in the End_status list,
					 * the value in the list must not start with '!' */
					DEBUG4( "Do_waitend: status '%s', End_status '%s'", status, End_status);
					if( (s = safestrstr( End_status, status ))
							&& (s == End_status || cval(s-1) != '!') ){
						waitend = 1;
					} else if( waitend_ctrl_t_interval > 0
							&& waitend_interval > waitend_ctrl_t_interval ){
						/* we wait only a short amount of time */
						waitend_interval = waitend_ctrl_t_interval;
					}
				}
			}
			DEBUG4("Do_waitend: waitend %d", waitend );
		}
	}
	if( waitend == 0 ){
		Errorcode = JFAIL;
		FATAL(_("Do_waitend: no end response from printer, timeout %d"), waitend_timeout );
	}
	LOGMSG(_("Do_waitend: end of %s detected"), banner?_("banner page"):_("job") );
}

/*
 * int Do_pagecount(void)
 * - get the pagecounter value from the printer
 *  We can use the PJL or PostScript method
 *  If we use PJL, then we send pagecount command, and read status
 *   until we get the value.
 *  If we use PostScript,  we need to set up the printer for PostScript,
 *   send the PostScript code,  and then get the response.
 *
 * INFO PAGECOUNT Option
 * pagecount=pjl   - uses PJL (default)
 * pagecount=ps    - uses PS
 * pagecount@
 *
 */

int Check_pagecount( int *use_ps_v, int *use_pjl_v )
{
	int use, use_ps, use_pjl;
	char *s;

	*use_ps_v = *use_pjl_v = 0;
	if( Status == 0 ) return(0);
	s = Pagecount;
	DEBUG4("Check_pagecount: status %d, pagecount using '%s'", Status, s );

	/* we see if we can do pagecount */
	use = use_pjl = use_ps = 0;
	if( !Is_flag( s, &use ) ){
		/* we check to see if the string is specifying */
		if( !strcasecmp(s,"pjl") ){
			use_pjl = 1;
		}
		if( !strcasecmp(s,"ps") ){
			use_ps = 1;
		}
		use = 1;
	} else if( use ){
		use_pjl = Pjl;
		use_ps = Ps;
	} else {
		return(0);
	}
	Strip_leading_spaces( &Ps_pagecount_code );
	if(use_ps && !Ps_pagecount_code){
		use_ps = 0;
	}
	if(use_pjl
		&& (Find_first_key( &Pjl_only, "info", 0, 0)
			||  !Find_first_key( &Pjl_except, "info", 0, 0)) ){
		use_pjl = 0;
	}
	if(use_ps && !Ps_pagecount_code){
		use_ps = 0;
	}

	if( !use_ps && !use_pjl ){
		Errorcode = JABORT;
		FATAL(_("Check_pagecount: pagecount '%s' and method not supported"), s );
	}
	if( use_pjl ) use_ps = 0;
	s = use_pjl?_("pjl info pagecount"):_("ps script");

	LOGMSG(_("Check_pagecount: pagecount using '%s'"), s );
	*use_ps_v = use_ps;
	*use_pjl_v = use_pjl;
	return(1);
}

int Do_pagecount( int pagecount_timeout, int pagecount_interval, int pagecount_poll,
	 int use_ps, int use_pjl )
{
	int pagecounter = 0, new_pagecounter = -1, len;
	int pagecount_same = 1;

	pagecounter = Current_pagecounter( pagecount_timeout,
		pagecount_interval, use_pjl, use_ps );
	if( pagecounter >= 0 && pagecount_poll > 0 ){
		LOGMSG(_("Do_pagecount: pagecounter %d, polling %d times at %d second intervals"),
			pagecounter, pagecount_poll, pagecount_interval );
		pagecount_same = 0;
		do{
			len = Read_status_timeout( pagecount_poll );
			DEBUG1("Do_pagecount: polling, getting pagecounter again");
			pagecounter = new_pagecounter;
			new_pagecounter = Current_pagecounter(pagecount_timeout,
				pagecount_interval, use_pjl, use_ps);
			if( new_pagecounter == pagecounter ){
				++pagecount_same;
			} else {
				pagecount_same = 0;
			}
		} while( len != -1
			&& (new_pagecounter == 0 || pagecount_same < pagecount_poll));
	}
	LOGMSG(_("Do_pagecount: pagecounter %d after %d attempts"),
		pagecounter, pagecount_same );
	return( pagecounter );
}

int Current_pagecounter( int pagecount_timeout, int pagecount_interval,
	int use_pjl, int use_ps )
{
	int i, len, flag, elapsed, timeout, page, pagecounter = 0;
	time_t start_t, current_t, interval_t;
	char *s;

	/* get start time */
	time( & start_t );

 again:

	if( Appsocket ){
		DEBUG1("Current_pagecounter: Appsocket device close");
		close(1);
		Open_device( Device );
		DEBUG1("Current_pagecounter: Appsocket device open done");
	}


	/* interval time */
	time( & interval_t );

	/*if(DEBUGL4)Dump_line_list("Current_pagecounter - Devstatus", &Devstatus);*/
	DEBUG1("Current_pagecounter: starting, use_pjl %d, use_ps %d, timeout %d",
		use_pjl, use_ps, pagecount_timeout );

	/* remove the old status */
	/* Set_str_value(&Devstatus,"pagecount",0); */

	Init_outbuf();
	if( use_pjl ){
		Put_outbuf_str( PJL_UEL_str );
		Put_outbuf_str( PJL_str );
		Put_pjl( PJL_INFO_PAGECOUNT_str );
		Put_outbuf_str( PJL_UEL_str );
	} else if( use_ps ){
		if( !Ps_pagecount_code ){
			Errorcode = JABORT;
			FATAL(_("Current_pagecounter: no ps_pagecount_code config info"));
		}
		if( Pjl ){
			Put_outbuf_str( PJL_UEL_str );
			if( Pjl_enter ){
				Put_outbuf_str( PJL_str );
				Put_outbuf_str( "@PJL ENTER LANGUAGE = POSTSCRIPT\n" );
			}
		}
		/* Watch out for ^D again */
		if( !Appsocket && Ps_eoj && Ps_eoj_at_start ) Put_outbuf_str( CTRL_D );
		Put_outbuf_str( Ps_pagecount_code );
		Put_outbuf_str( "\n" );
		if( !Appsocket && Ps_eoj && Ps_eoj_at_end ) Put_outbuf_str( CTRL_D );
		if( Pjl ) Put_outbuf_str( PJL_UEL_str );
	}

	DEBUG2("Current_pagecounter: using '%s'", Outbuf );

	page = 0;
	while(!page){
		flag = 0;
		/* write the string */
		/* check to see if we need to read */
		time( &current_t );
		elapsed = current_t - start_t;
		if( pagecount_timeout > 0 ){
			if( elapsed >= pagecount_timeout ){
				break;
			}
			timeout = pagecount_timeout - elapsed;
		} else {
			timeout = 0;
		}
		if( pagecount_interval > 0 ){
			elapsed = current_t - interval_t;
			if( elapsed > pagecount_interval ){
				goto again;
			}
			if( timeout == 0 || timeout > pagecount_interval ){
				timeout = pagecount_interval;
			}
		}
		if( Outlen ){
			len = Write_read_timeout( Outlen, Outbuf, timeout );
			Init_outbuf();
		} else {
			len = Read_status_timeout( timeout );
		}
		DEBUG1("Current_pagecounter: write len %d", len );
		if( len == -1 ){
			break;
		}
		if(DEBUGL4)Dump_line_list("Current_pagecounter - Devstatus",&Devstatus);
		if( !Find_first_key( &Devstatus, "pagecount", Value_sep, &i )){
			s = Find_str_value( &Devstatus, "pagecount", Value_sep );
			if( s && isdigit(cval(s)) ){
				pagecounter = atoi( s );
				page = 1;
			}
		}
	}
	DEBUG1("Current_pagecounter: page %d, pagecounter %d", page, pagecounter );
	if( page == 0 ){
		Errorcode = JFAIL;
		FATAL(_("Current_pagecounter: no pagecounter response from printer, timeout %d"), pagecount_timeout );
	}
	return( pagecounter );
}


/*
 * Send_job
 * 1. we read an input buffer
 * 2. we then check to see what type it is
 * 3. we then set up the various configuration
 * 4. we then send the job to the remote end
 * Note: there is a problem when you send language switch commands
 *  to the HP 4M+, and other printers.
 * two functions.
 *
 * The following comments from hp.com try to explain the situation:
 *
 *"What is happening is the printer buffer fills with your switch
 * command and the first parts of your job. The printer does not
 * "prescan" the input looking for the <Esc>%-1234... but instead gets
 * around to it after some time has expired. During the switch, the
 * data that you have already trasmitted is discarded and is picked
 * up whereever your driver happens to be. For PostScript that's a
 * messy situation.
 * 
 * What you need to do is isolate the switch command. You can do this
 * with nulls if you have no control over the timing. 8K of them will
 * fill the largest allocated I/O buffer and then hold off the UNIX
 * system. The switch will be made and the initial remaining nulls
 * will be discarded. If you can control the timing of the data, you'll
 * have to experiment with the correct time.
 * 
 */

 char *PJL_ENTER_str = "@PJL ENTER LANGUAGE = %s\n";
 char *PCL_EXIT_str = "\033E";

void Send_job()
{
	int len = 0, i, c, n, cnt, tempfd;
	char *s, *pgm = 0, *t;
	int done = 0;
	struct line_list l, match, files;
	char *language;
	char *file_util;
	struct stat statb;
	int progress_pc, progress_k;
	double total_size, progress_last, progress_total;
	char file_result[SMALLBUFFER];
	char errmsg[SMALLBUFFER];
	int error_count = 0;

	Init_line_list( &l );
	Init_line_list( &match );
	Init_line_list( &files );

	LOGMSG( _("Send_job: starting transfer") );

	Make_stdin_file();
	file_util = Find_str_value( &Model, "file_util_path", Value_sep );
	file_util = Fix_option_str( file_util, 0, 1, 1 );

	/* find if there are user specified langagues */
	language = 0;
 	if( !Force_processing && (Loweropts['c'-'a'] || Autodetect) ){
		language = RAW;
	} else if( (s = Find_str_value(&Zopts,"language",Value_sep)) && !ISNULL(s) ){
		language = s;
	}

 again:

	if( pgm ) free(pgm); pgm = 0;
	Init_outbuf();
	if( lseek(0,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Send_job: lseek failed"));
	}
	/* we only need the first couple of characters */
	cnt = 64;
	if( cnt > Outmax ) cnt = Outmax; 
	DEBUG2( "Send_job: want %d", cnt );
	Outlen = read( 0, Outbuf, cnt);
	Outbuf[Outlen] = 0;
	DEBUG2( "Send_job: read %d from stdin '%s'", Outlen, Outbuf );
	if( Outlen < 0 ){
		Errorcode = JFAIL;
		LOGERR_DIE( _("Send_job: read error on stdin") );
	}
	if( Outlen == 0 ){
		LOGMSG( _("Send_job: zero length job file") );
		return;
	}
	if( lseek(0,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Send_job: lseek failed"));
	}

	/* decide on the job type */

	Set_str_value( &Model,"file_output","");
	if( language ){
		/* we have the language already */
	} else if( Force_conversion ){
		if( ISNULL(file_util) ){
			Errorcode = JABORT;
			FATAL( _("Send_job: missing file_util_path value"));
		}
		DEBUG2( "Send_job: file_util_path '%s'", file_util );
		Use_file_util(file_util, file_result, sizeof(file_result) );
		language = file_result;
	} else if( !strncmp( PJL_UEL_str, Outbuf, strlen(PJL_UEL_str)) ){
		language = PJL;
	} else if( Outbuf[0] == '\033' ){
		language = PCL;
	} else if( Outbuf[0] == '\004' && !strncmp(Outbuf+1, "%!", 2) ){
		language = PS;
	} else if( !strncmp(Outbuf, "%!", 2) ){
		language = PS;
	} else if( !ISNULL(file_util) ){
		DEBUG2( "Send_job: file_util_path '%s'", file_util );
		file_result[0] = 0;
		Use_file_util(file_util, file_result, sizeof(file_result) );
		if( !ISNULL(file_result) ) language = file_result;
	}
	if( ISNULL(language) ){
		/* defaults */
		language = Find_str_value(&Model,"default_language",Value_sep);
	}
	if( ISNULL(language) ) language = UNKNOWN;

	/* rewind the file and get ready for action */
	Init_outbuf();
	if( lseek(0,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Send_job: lseek failed"));
	}

	LOGMSG( _("Send_job: initial job type '%s'"), language );

	if( safestrcmp(language,RAW)
		&& (s = Find_str_value(&Model, "file_output_match",Value_sep)) ){
		/* split the matches up */
		Free_line_list(&l);
		Split(&match,s,Line_ends,0,0,0,0,0);
		if( match.count > 0 && (match.list[0])[0] == '<' ){
			Split(&files,(match.list[0])+1,Value_sep,0,0,0,0, 0 );
			for( i = 0; i < files.count; ++i ){
				s = files.list[i];
				DEBUG2("Send_job: reading file '%s'", s );
				t = Get_file_image( s );
				if( ISNULL(t) ){
					Errorcode = JABORT;
					LOGERR_DIE(_("Send_job: could not open '%s'"), s);
				}
				DEBUG2("Send_job: file '%s'", t );
				Split(&match,t,Line_ends,0,0,0,0, 0 );
				if( t ) free(t); t = 0;
			}
			Free_line_list(&files);
			if(DEBUGL2)Dump_line_list("Send_job: after files", &match );
		}
		for( i = 0; i < match.count; ++i ){
			s = match.list[i];
			DEBUG2("Send_job: [%d] '%s'", i, s );
			if( !s ) continue;
			while( isspace(cval(s)) || cval(s) == ']' || cval(s) == '[' ) ++s;
			Free_line_list(&l);
			Split_count(2,&l,s,Whitespace, 0, 0, 0, 1, 1 );
			if(DEBUGL2)Dump_line_list("Send_job: checking against", &l );
			if( l.count == 0 ) continue;
			s = l.list[0];
			if( l.count != 2 && l.count != 3 ){
				LOGMSG( _("Send_job: wrong number of fields in 'file_output_match' - '%s'"), s );
				continue;
			}
			if( !Globmatch( s, language ) ){
				SNPRINTF(file_result,sizeof(file_result)) l.list[1]);
				language = file_result;
				if( l.count == 3 ){
					pgm = Fix_option_str( l.list[2], 0, 1, 1 );
				}
				break;
			}
		}
		Free_line_list(&l);
	}

	language = Set_mode_lang(language);
	LOGMSG( _("Send_job: decoded job type '%s'"), language );

	/* we can treat text as PCL */
	if( !safestrcasecmp(language,RAW) ){
		;
	} else if( !safestrcasecmp(language,MSG) ){
		LOGMSG( _("Send_job: MSG '%s'"), language );
	} else if( !safestrcasecmp(language,PJL) ){
		if( !Pjl ){
			Errorcode = JABORT;
			FATAL(_("Send_job: job is PJL and PJL not supported"));
		}
	} else if( !safestrcasecmp(language,PCL) ){
		if(!Pcl ){
			Errorcode = JABORT;
			FATAL(_("Send_job: job is PCL and PCL not supported"));
		}
	} else if( !safestrcasecmp(language,PS) ){
		if( !Ps ){
			Errorcode = JABORT;
			FATAL(_("Send_job: job is PostScript and PostScript not supported"));
		}
	} else if( !safestrcasecmp(language,TEXT) ){
		if( !Pcl && !Text ){
			Errorcode = JABORT;
			FATAL(_("Send_job: job is Text and Text not supported"));
		}
		if( Pcl ) language = PCL;
	} else if( !safestrcasecmp(language,FILTER) ){
		if( ISNULL(pgm) ){
			Errorcode = JABORT;
			FATAL(_("Send_job: filter wanted and no filter specified"));
		}
	} else {
		LOGMSG( _("Send_job: cannot process language '%s'"), language );
		return;
	}

	/* we now invoke the converter or do hanky-panky with the MSG
	 * this is called a hack of the 2nd order... Patrick
	 */
	if( !ISNULL( pgm ) ){
		LOGMSG( _("Send_job: job type '%s', converter '%s'"),
			language, pgm );
		if( lseek(0,0,SEEK_SET) == -1 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Send_job: lseek failed"));
		}
		tempfd = Make_tempfile();
		/* here is where we do the dirty work */
		if( !safestrcasecmp( language, MSG ) ){
			/* now we cunningly force a rescan
			 * gahhh... I put this in but I hoped I would never
			 * have to use it
			 */
			Write_error_msg( tempfd, pgm );
			language = FILTER;
		} else {
			n = Filter_file( pgm, _("CONVERTER"), 0, tempfd, errmsg, sizeof(errmsg) );
			if( n ){
				close(tempfd);
				tempfd = Make_tempfile();
				/* stop recursion from causing problems */
				if( error_count > 2 ) {
					Errorcode = n;
					FATAL(_("Send_job: converter failed, exit code %d"), n);
				}
				++error_count;
				Set_str_value( &Model,"pgm",pgm);
				Set_str_value( &Model,"msg",errmsg);
				Write_error_msg( tempfd, "conversion_error" );
				language = FILTER;
			}
		}
		if( fstat(tempfd,&statb) == -1 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Send_job: fstat2 failed"));
		}
		LOGMSG(_("Send_job: converter done, output %ld bytes"),
			(long)(statb.st_size) );
		if( statb.st_size == 0 ){
			Errorcode = JABORT;
			LOGMSG(_("Send_job: zero length conversion output"));
		}
		if( tempfd != 0 ){
			if( dup2(tempfd,0) == -1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Send_job: dup2 failed"));
			}
			if( close(tempfd) == -1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Send_job: close(tempfd %d) failed"), tempfd);
			}
		}
		/* if you did filtering,  then reopen for the job */
		if( pgm ) free(pgm); pgm = 0;
	} else if( !safestrcmp(language,PJL) && Remove_pjl_at_start ){
		LOGMSG( _("Send_job: job type '%s', stripping leading PJL "), language  );
		if( lseek(0,0,SEEK_SET) == -1 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Send_job: lseek failed"));
		}
		tempfd = Make_tempfile();
		Init_outbuf();
		do{
			len = read( 0, Outbuf+Outlen, Outmax - Outlen);
			DEBUG2( "Send_job: read %d from stdin", len );
			if( len < 0 ){
				Errorcode = JFAIL;
				LOGERR_DIE( _("Send_job: read error on stdin") );
			}
			Outlen += len;
			Outbuf[Outlen] = 0;
			done = 0;
			while(!done){
				n = strlen(PJL_UEL_str);
				if( !strncmp( PJL_UEL_str, Outbuf, n) ){
					memmove( Outbuf, Outbuf+n, Outlen - n + 1);
					Outlen -= n;
					continue;
				} else if( !strncmp(Outbuf,"@PJL",4) ){
					if( (s = safestrpbrk(Outbuf,"\n")) ){
						*s++ = 0;
						n = s - Outbuf;
						memmove( Outbuf, s, Outlen - n + 1);
						Outlen -= n;
						continue;
					} else {
						break;
					}
				}
				done = 1;
			}
		} while( !done && len > 0 );
		/* now we write this to the output file */
		/*DEBUG2("Send_job: initial part len %d, '%s'", Outlen, Outbuf ); */
		do{
			if( Outlen > 0 ){
				len = write( tempfd, Outbuf, Outlen );
				if( len != Outlen ){
					Errorcode = JFAIL;
					LOGERR_DIE( _("Send_job: write error to tempfile") );
				}
			}
			Outlen = read( 0, Outbuf, Outmax );
			Outbuf[Outlen] = 0;
			/* DEBUG2("Send_job: next part len %d, '%s'", Outlen, Outbuf ); */
		} while( Outlen > 0 );
		if( dup2(tempfd,0) == -1 ){
			Errorcode = JABORT;
			LOGERR_DIE( _("Make_stdin_file: dup2 failed") );
		}
		if( tempfd != 0 ) close(tempfd);
		language = 0;
		goto again;
	} else {
		LOGMSG( _("Send_job: job type '%s'"), language  );
	}

	/* run the job through the converter */
	if( !safestrcmp(language,FILTER) ){
		language = 0;
		goto again;
	}

	Init_outbuf();
	Init_job( language );
	Init_outbuf();

	DEBUG1("Send_job: doing file transfer");
	if( fstat( 0, &statb ) < 0 ){
		Errorcode = JABORT;
		LOGERR_DIE( _("Send_job: cannot fstat fd 0") );
	}
	total_size = statb.st_size;
	if( lseek(0,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Send_job: lseek failed"));
	}

	progress_last = progress_total = 0;
	progress_pc = progress_k = 0;
	if( (s = Find_str_value(&Model,"progress_pc",Value_sep)) ){
		progress_pc = atoi(s);
	}
	if( (s = Find_str_value(&Model,"progress_k",Value_sep)) ){
		progress_k = atoi(s);
	}
	if( progress_pc < 0 || progress_pc > 100 ){
		progress_pc = 0;
	}
	if( progress_k <= 0 ) progress_k = 0;
	if( progress_pc == 0 && progress_k == 0 ) progress_k = 100;
	DEBUG2( "Send_job: input stat 0%06o, size %0.0f, progress_pc %d, progress_k %d",
		(int)(statb.st_mode & S_IFMT),
		total_size, progress_pc, progress_k );
	LOGMSG( _("Send_job: transferring %0.0f bytes"), total_size );

	if( !safestrcmp(language,PS) && Remove_ctrl ){
		for( i = 0; i < strlen(Remove_ctrl); ++i ){
			Remove_ctrl[i] &= 0x1F;
		}
	}

	Init_outbuf();

	/* just a couple of bytes at the start */
	Outlen = read( 0, Outbuf, 8);
	Outbuf[Outlen] = 0;
	DEBUG2( "Send_job: read %d from stdin", Outlen );
	if( Outlen < 0 ){
		Errorcode = JFAIL;
		LOGERR_DIE( _("Send_job: read error on stdin") );
	}
	if( Outlen == 0 ){
		LOGMSG( _("Send_job: zero length job file") );
		return;
	}

	/* strip off the ^D at start of PostScript jobs */
	if( !safestrcmp(language,PS) && Outbuf[0] == '\004' ){
		memmove(Outbuf,Outbuf+1,Outlen-1);
		Outlen -= 1;
	}
	if( Outlen > 1 && !safestrcmp(language,PCL) && Outbuf[0] == '\033' && Outbuf[1] == 'E' ){
		memmove(Outbuf,Outbuf+2,Outlen-2);
		Outlen -= 2;
	}
	Outbuf[Outlen] = 0;
	DEBUG2( "Send_job: Outbuf after type and stripping '%s'", Outbuf );

	done = 0;
	do{
		len = 0;
		progress_total += Outlen;
		if( !safestrcmp(language,PS) && Tbcp ){
			DEBUG4("Send_job: tbcp");
			for( cnt = 0; cnt < Outlen; ){
				len = 0; c = 0;
				for( i = cnt; c == 0 && i < Outlen; ){
					switch( (c = ((unsigned char *)Outbuf)[i]) ){
					case 0x01: case 0x03: case 0x04: case 0x05:
					case 0x11: case 0x13: case 0x14: case 0x1C:
						break;
					default: c = 0; ++i;
						break;
					}
				}
				/* we now write the string from cnt to count+i-1 */
				n = i - cnt;
				DEBUG1("Send_job: tbcp writing %d", n );
				if( (len = Write_read_timeout( n, Outbuf+cnt, Job_timeout )) ){
					break;
				}
				if( c ){
					char b[2];
					DEBUG1("Send_job: tbcp escape 0x%02x", c );
					b[0] = '\001'; b[1] = c ^ 0x40;
					if( (len = Write_read_timeout( 2, b, Job_timeout )) ){
						break;
					}
					n += 1;
				}
				cnt += n;
			}
		} else if( !safestrcmp(language,PS) && Remove_ctrl ){
			for( i = 0; i < Outlen; ){
				c = cval(Outbuf+i);
				if( safestrchr( Remove_ctrl,c ) ){
					memmove( Outbuf+i, Outbuf+i+1,Outlen-i);
					--Outlen;
				} else {
					++i;
				}
			}
			len = Write_read_timeout( Outlen, Outbuf, Job_timeout );
		} else if(
			(!safestrcmp(language,TEXT) || !safestrcmp(language,PCL))
				&& Crlf ){
			DEBUG4("Send_job: crlf");
			for( cnt = 0; cnt < Outlen; ){
				len = 0; c = 0;
				for( i = cnt; c == 0 && i < Outlen; ){
					switch( (c = ((unsigned char *)Outbuf)[i]) ){
					case '\n':
						break;
					default: c = 0; ++i;
						break;
					}
				}
				/* we now write the string from cnt to count+i-1 */
				n = i - cnt;
				DEBUG1("Send_job: crlf writing %d", n );
				if( (len = Write_read_timeout( n, Outbuf+cnt, Job_timeout )) ){
					break;
				}
				if( c ){
					if( (len = Write_read_timeout( 2, "\r\n", Job_timeout )) ){
						break;
					}
					n += 1;
				}
				cnt += n;
			}
		} else {
			DEBUG1("Send_job: writing %d", Outlen );
			len = Write_read_timeout( Outlen, Outbuf, Job_timeout );
		}
		if( len ){
			Errorcode = JFAIL;
			LOGERR_DIE( _("Send_job: job failed during copy") );
		}
		Init_outbuf();
		DEBUG1("Send_job: total written %0.0f", progress_total );
		/*
		 * now we do progress report
		 */
		i = (progress_total - progress_last)/total_size*100;
		if( progress_pc && (i >= progress_pc) ){
			DEBUG1("Send_job: pc total %0.0f, change %d%%", progress_total, i );
			i = progress_total/total_size*100;
			LOGMSG(_("Send_job: %d percent done"), i);
			progress_last = progress_total;
		} else if( progress_k ){
			i = (progress_total - progress_last)/1024;
			DEBUG1("Send_job: k total %0.0f, change %d", progress_total, i );
			if( i >= progress_k ){
				if( progress_pc ){
					i = progress_total/total_size*100;
					LOGMSG(_("Send_job: %d percent done"), i);
				} else
					LOGMSG(_("Send_job: %d Kbytes done"), (int)(progress_total/1024));
				progress_last = progress_total;
			}
		}
		if( Outlen == 0 && !done ){
				DEBUG1("Send_job: read %d", len );
			while( Outlen < Outmax 
					&& (len = read( 0, Outbuf+Outlen, Outmax - Outlen )) > 0 ){
				DEBUG1("Send_job: read %d", len );
				Outlen += len;
			}
			Outbuf[Outlen] = 0;
			DEBUG1("Send_job: read total %d", Outlen );
			if( len < 0 ){
				Errorcode = JFAIL;
				LOGERR_DIE( _("Send_job: read error on stdin") );
			} else if( len == 0 ){
				done = 1;
			}
		}
	} while( Outlen > 0 );
	DEBUG1( "Send_job: finished file transfer" );

	Term_job( language );

	Free_line_list(&l);
	Free_line_list(&files);
	Free_line_list(&match);
	LOGMSG( _("Send_job: data sent") );
}

void URL_decode( char *s )
{
	if( s ) while( (s = safestrpbrk(s, "%?")) ){
		if( cval(s) == '?' ){
			*s = ' ';
		} if( (s[0] = s[1]) && (s[1] = s[2]) ){
			s[2] = 0;
			s[0] = strtol(s,0,16);
			memmove(s+1,s+3, strlen(s+3)+1);
			++s;
		} else {
			break;
		}
	}
}

/*
 * Builtins
 */

 struct keyvalue Builtin_values[] = {
/* we get the various UEL strings */
{"Font_download","font_download",(void *)Font_download},
{0,0,0}
};


/*
 * Process_OF_mode()
 *  We read from STDIN on a character by character basis.
 *  We then send the output to STDOUT until we are done
 *  If we get a STOP sequence, we suspend ourselves.
 *
 *  We actually have to wait on 3 FD's -
 *  0 to see if we have input
 *  1 to see if we have output
 *
 *  We will put the output in the Outbuf buffer
 *  Note:  if we need to suspend,  we do this AFTER we have
 *   written the output
 */

 static char stop[] = "\031\001";    /* sent to cause filter to suspend */

int Process_OF_mode(void)
{
	char inbuf[2];
	int n, state;
	n = state = 0;
	if( ftruncate( 1, 0 ) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Process_OF_mode: ftruncate fd 1 failed"));
	}
	while( (n = read(0,inbuf,1)) > 0 ){
		if( inbuf[0] == stop[state] ){
			++state;
			if( stop[state] == 0 ){
				return(1);
			}
			continue;
		} else if( state ){
			state = 0;
			if( write(1,stop,1) != 1 ){
				Errorcode = JABORT;
				LOGERR_DIE(_("Process_OF_mode: write failed"));
			}
		}
		if( write(1,inbuf,1) != 1 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Process_OF_mode: write failed"));
		}
	}
	if( state && write(1,stop,1) != 1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Process_OF_mode: write failed"));
	}
	return(n);
}

void Add_val_to_list( struct line_list *v, int c, char *key, char *sep )
{
	char *s = 0;
	if( isupper(c) ) s = Upperopts[c-'A'];
	if( islower(c) ) s = Loweropts[c-'a'];
	if( s ){
		s = safestrdup3(key,sep,s,__FILE__,__LINE__);
		DEBUG3("Add_val_to_list: '%s'",s );
		Add_line_list( v, s, sep, 1, 1 );
		free(s); s = 0;
	}
}


/*
 * set the close on exec flag for a reasonable range of FD's
 */
void close_on_exec( int n )
{
    int fd, max = Max_fd+10;

    for( fd = n; fd < max; fd++ ){
		fcntl(fd, F_SETFD, 1);
	}
}

/*
 * Use_file_util(char *pgm, char *matches, struct line_list *args
 * 
 * We will need to use the file utility
 *  pgm = program to invoke
 *  matches = lines in form of 
 *      glob   returntype  program
 *  args = arguments to use if successful
 * RETURN 0 if no match
 *        != 0 if match
 */

void Use_file_util(char *pgm, char *value, int value_len )
{
	int n, fd, c;
	char *s, *t;

	DEBUG1("Use_file_util: pgm '%s', value_len %d", pgm, value_len);
	LOGMSG(_("Use_file_util: file program = '%s'"), pgm );

	value[0] = 0;
	if( lseek(0,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Use_file_util: lseek failed"));
	}
	fd = Make_tempfile();
	n = Filter_file( pgm, _("FILE_UTIL"), 0, fd, 0, 0 );
	if( n ){
		Errorcode = n;
		LOGERR_DIE(_("Use_file_util: exit code %d"), n );
	}
	if( lseek(fd,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Use_file_util: lseek failed"));
	}
	if( (n = read(fd,value,value_len-1)) < 0){
		Errorcode = JABORT;
		LOGERR_DIE(_("Use_file_util: read failed"));
	}
	if( n >= 0 ) value[n] = 0;
	close(fd);
	DEBUG1("Use_file_util: read %d, '%s'", n, value );
	lowercase(value);
	if( (s = safestrchr(value,':')) ){
		++s;
	} else {
		s = value;
	}
	while( isspace(cval(s)) ) ++s;
	for( t = value; (c = cval(s)) ; ++s ){
		if( isspace(cval(s)) && (isspace(cval(s+1)) || !cval(s+1)) ){
			;
		} else if( isspace(cval(s)) ){
			*t++ = '_';
		} else {
			*t++ = *s;
		}
	}
	*t = 0;
	LOGMSG(_("Use_file_util: file information = '%s'"), value );
	DEBUG4("Use_file_util: file util done, '%s'", value );
	Set_str_value( &Model,"file_output",value);
}

int Make_tempfile( void )
{
	int fd;
	char buffer[32];
	char *tempfile, *t;

	tempfile = Find_str_value( &Model,"tempfile", Value_sep );
	if( tempfile && *tempfile == 0 ) tempfile = 0;
	if( tempfile == 0 ) tempfile = getenv("TMP");
	if( tempfile && *tempfile == 0 ) tempfile = 0;
	SNPRINTF(buffer,sizeof(buffer)) "ifhpXXXXXX");
	t = safestrdup3(tempfile,tempfile?"/":"",buffer,__FILE__,__LINE__);
	DEBUG2( "Make_tempfile: tempfile '%s'", t );
	if( (fd = mkstemp(t)) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE("Make_tempfile: mkstemp failed with '%s'", t);
	}
	DEBUG2( "Make_tempfile: new tempfile '%s', fd %d", t, fd );
	unlink(t);
	free(t); t = 0;
	return( fd );
}

void Make_stdin_file()
{
	int is_file, fd, in, n, i;
	struct stat statb;
	char *s;

	if( fstat( 0, &statb ) < 0 ){
		Errorcode = JABORT;
		LOGERR_DIE( _("Make_stdin_file: cannot fstat fd 1") );
	}
	is_file = ((statb.st_mode & S_IFMT) == S_IFREG);
	DEBUG2( "Make_stdin_file: input is_file %d, size %d",
		is_file, (int)statb.st_size );
	if( !is_file ){
		fd = Make_tempfile();
		in = 1;
		do{
			n = 0;
			for( i = Outlen, s = Outbuf;
				i > 0 && (n = write(fd,s,i)) > 0;
				i -= n, s += n);
			if( n < 0 ){
				Errorcode = JABORT;
				LOGERR_DIE( _("Make_stdin_file: write to tempfile fd %d failed"), fd );
			}
			Outlen = 0;
			if( in > 0 ){
				n = 0;
				for( Outlen = 0;
					Outlen < Outmax
					&& (n = read(0,Outbuf+Outlen,Outmax-Outlen)) > 0;
					Outlen+=n);
				if( n < 0 ){
					Errorcode = JABORT;
					LOGERR_DIE( _("Make_stdin_file: read from stdin") );
				} else if( n == 0 ){
					in = 0;
				}
			}
		} while( Outlen > 0 );
		if( dup2(fd,0) == -1 ){
			Errorcode = JABORT;
			LOGERR_DIE( _("Make_stdin_file: dup2 failed") );
		}
		if( fd != 0 ) close(fd);
	}
	if( lseek(0,0,SEEK_SET) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Make_stdin_file: lseek failed"));
	}
	Outlen = 0;
}

char * Set_mode_lang( char *s )
{
	if( !ISNULL(s) ){
		if(!strcasecmp( s, PS )){ s = PS; }
		else if(!strcasecmp( s, "ps" )){ s = PS; }
		else if(!strcasecmp( s, "postscript" )){ s = PS; }
		else if(!strcasecmp( s, PCL )){ s = PCL; }
		else if(!strcasecmp( s, PJL )){ s = PJL; }
		else if(!strcasecmp( s, TEXT )){ s = TEXT; }
		else if(!strcasecmp( s, RAW )){ s = RAW; }
		else if(!strcasecmp( s, FILTER )){ s = FILTER; }
	}
	return s;
}

/* Fd_readable - check to see if you can read status from the file descriptor
 * use fcntl and stat to make sure that you do not have
 *   a file as the device and that it is open RDWR
 */

int Fd_readable( int fd, int *poll_for_status )
{
	int nonblock, n, readable = 0;
	struct stat statb;
	char buffer[1];


	if( fstat(fd,&statb) == -1 ){
		Errorcode = JABORT;
		FATAL( _("ifhp: fd is not open"),fd);
	}
	if( S_ISREG( statb.st_mode ) ){
		readable = 0;
		return( readable );
	}
	nonblock = Get_nonblock_io( fd );
	if( !nonblock ){
		Set_nonblock_io( fd );
	}
	errno = 0;
	n = read( fd, buffer, 1 );
	DEBUG1("Fd_readable: read %d, errno %d (%s)", n, errno, strerror(errno)   );
	if( n >= 0 || errno == EWOULDBLOCK ){
		readable = 1;
		if( !isatty( fd ) ){
			if( fstat( fd, &statb ) < 0 ){
				Errorcode = JABORT;
				LOGERR_DIE( _("Fd_readable: cannot fstat fd %d"), fd );
			}
			DEBUG2( "Fd_readable: fd %d st_mode 0%o", fd, statb.st_mode & S_IFMT );
			if ( S_ISCHR(statb.st_mode ) ){
				*poll_for_status = 1;
			}
		}
	}
	if( !nonblock ){
		Set_block_io( fd );
	}
	return( readable );
}

void Init_job( char *language )
{
	char buffer[SMALLBUFFER];
	struct line_list l;
	int len, i;
	char *s;

	Init_line_list(&l);

	DEBUG2("Init_job: language '%s'", language );
	if( !safestrcmp(language,TEXT) ) language = PCL;
	if( Pjl && Pjl_enter && (!safestrcmp(language,PS) || !safestrcmp(language,PCL)) ){
		SNPRINTF(buffer, sizeof(buffer)) 
			PJL_ENTER_str, language );
		Put_pjl( buffer );
		memset( buffer, 0, sizeof(buffer) );
		/* 
		 * force nulls to be output
		 */
		for( i = Null_pad_count; i > 0; i -= len){
			len = sizeof(buffer);
			if( i < len ) len = i;
			Put_outbuf_len( buffer, len );
		}
	}
	DEBUG2("Init_job: after PJL ENTER '%s'", Outbuf );

	/* now we handle the various initializations */
	if( Pcl && !safestrcmp(language,PCL) ){
		DEBUG1("Init_job: doing pcl init");
		if( Pcl_eoj_at_start ) Put_outbuf_str( PCL_EXIT_str );
		if( (s = Find_str_value( &Model, "pcl_init", Value_sep)) ){
			DEBUG1("Init_job: 'pcl_init'='%s'", s);
			Free_line_list( &l );
			Split( &l, s, List_sep, 1, 0, 1, 1, 0 );
			Resolve_list( "pcl_", &l, Put_pcl );
			Free_line_list( &l );
		}
		DEBUG1("Init_job: 'pcl' and Topts");
		Resolve_user_opts( "pcl_", &Pcl_user_opts, &Unsorted_Topts, Put_pcl );
		DEBUG1("Init_job: 'pcl' and Zopts");
		Resolve_user_opts( "pcl_", &Pcl_user_opts, &Unsorted_Zopts, Put_pcl );
	} else if( Ps && !safestrcmp(language,PS) ){
		DEBUG1("Init_job: doing ps init");
		if( !Appsocket && Ps_eoj && Ps_eoj_at_start ) Put_outbuf_str( CTRL_D );
		if( Tbcp ){
			DEBUG2("Init_job: doing TBCP");
			Put_outbuf_str("\001M");
		}
		if( (s = Find_str_value(&Model, "ps_level_str",Value_sep)) ){
			Put_ps(s);
		} else {
			Put_ps("%!");
		}
		if( (s  = Find_str_value( &Model, "ps_init", Value_sep)) ){
			DEBUG1("Init_job: 'ps_init'='%s'", s);
			Free_line_list( &l );
			Split( &l, s, List_sep, 1, 0, 1, 1, 0 );
			Resolve_list( "ps_", &l, Put_ps );
			Free_line_list( &l );
		}
		DEBUG1("Init_job: 'ps' and Topts");
		Resolve_user_opts( "ps_", &Ps_user_opts, &Unsorted_Topts, Put_ps );
		DEBUG1("Init_job: 'ps' and Zopts");
		Resolve_user_opts( "ps_", &Ps_user_opts, &Unsorted_Zopts, Put_ps );
	}
	DEBUG2("Init_job: final '%s'", Outbuf );
	len = Write_read_timeout( Outlen, Outbuf, Job_timeout );
	if( len ){
		Errorcode = JFAIL;
		FATAL( "Init_job: job timed out language selection" );
	}
	Init_outbuf();
	Free_line_list(&l);
}

void Term_job( char *language )
{
	int len;
	struct line_list l;
	char *s;

	DEBUG2("Term_job: language '%s'", language );

	Init_line_list(&l);
	/* now we handle the various terminations */
	if( Pcl && !safestrcmp(language,PCL) ){
		DEBUG1("Term_job: doing pcl term");
		if( (s = Find_str_value( &Model, "pcl_term", Value_sep)) ){
			DEBUG1("Term_job: 'pcl_term'='%s'", s);
			Free_line_list( &l );
			Split( &l, s, List_sep, 1, 0, 1, 1, 0 );
			Resolve_list( "pcl_", &l, Put_pcl );
			Free_line_list( &l );
		}
	} else if( Ps && !safestrcmp(language,PS) ){
		DEBUG1("Term_job: doing ps term");
		if( (s = Find_str_value( &Model, "ps_term", Value_sep)) ){
			DEBUG1("Term_job: 'ps_term'='%s'", s);
			Free_line_list( &l );
			Split( &l, s, List_sep, 1, 0, 1, 1, 0 );
			Resolve_list( "ps_", &l, Put_ps );
			Free_line_list( &l );
		}
	}
	if( Pcl && ( !safestrcmp(language,PCL) || !safestrcmp(language,TEXT) ) ){
		Put_outbuf_str( PCL_EXIT_str );
	}
	if( Ps && !safestrcmp(language,PS) ){
		if( Ps_eoj && Ps_eoj_at_end ) Put_outbuf_str( CTRL_D );
	}
	len = Write_read_timeout( Outlen, Outbuf, Job_timeout );
	if( len ){
		Errorcode = JFAIL;
		FATAL( _("Term_job: job error on cleanup strings") );
	}
	Init_outbuf();
	Free_line_list(&l);
}

/*
 * filter a file through a program
 */

int Filter_file( char *pgm, char *title, int fd_stdin, int fd_stdout,
	char *errbuffer, int len_errbuffer )
{
	char *s;
	struct line_list l;
	int error_fd[2], pid, len, n;
	char buffer[SMALLBUFFER];
	plp_status_t status;

	Init_line_list(&l);
	if(cval(pgm) == '|') ++pgm;
	if( safestrstr(pgm,"ZOPTS") || safestrstr(pgm,"TOPTS") ){
		LOGMSG(_("Filter_file: WARNING - 'ZOPTS' and 'TOPS' conversion options replaced by {Z} and {T} - '%s'"), pgm ); 
	}
	if( (s = safestrstr(pgm,"ARGV")) && s[4] != '}' ){
		LOGMSG("Filter_file: WARNING - 'ARGV' conversion option replaced by {ARGV} - '%s'", pgm ); 
	}
	pgm = safestrdup(pgm, __FILE__, __LINE__ );
	DEBUG4("Filter_file: pgm '%s', title '%s', stdin %d, stdout %d",
		pgm, title, fd_stdin, fd_stdout );
	if( pgm && (safestrpbrk( pgm, "|><;" ) || (cval(pgm) == '(')) ){
		Split(&l, "/bin/sh -c", Whitespace, 0, 0, 0, 1, 0 );
		Check_max(&l,2);
		if( cval(pgm) != '(' ){
			l.list[l.count++] = safestrdup3("( ",pgm," )", __FILE__, __LINE__ );
		} else {
			l.list[l.count++] = safestrdup(pgm, __FILE__, __LINE__ );
		}
		l.list[l.count] = 0;
	} else {
		Split_cmd_line( &l, pgm );
		Check_max(&l,2);
		l.list[l.count] = 0;
	}
	if(DEBUGL2)Dump_line_list("Filter_file: process args", &l );
	if( pipe(error_fd) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Filter_file: pipe failed"));
	}
	Set_max_fd(error_fd[0]); Set_max_fd(error_fd[1]);

	/* fork the process */
	if( (pid = fork()) == -1 ){
		Errorcode = JABORT;
		LOGERR_DIE(_("Filter_file: fork failed"));
	} else if( pid == 0 ){
		if( fd_stdin < 0 && (fd_stdin = open("/dev/null", O_RDWR)) < 0 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Filter_file: open(/dev/null) failed"));
		}
		if( fd_stdout < 0 && (fd_stdout = open("/dev/null", O_RDWR)) < 0 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Filter_file: open(/dev/null) failed"));
		}
		if( (fd_stdin != 0 && dup2(fd_stdin,0) == -1)
			|| (fd_stdout != 1 && dup2(fd_stdout,1) == -1)
			|| dup2(error_fd[1],2) == -1 ){
			Errorcode = JABORT;
			LOGERR_DIE(_("Filter_file: dup2 failed"));
		}
		close_on_exec(3);
		/* now we exec the filter process */
		execve( l.list[0], l.list, Envp );
		/* ooops... error */
		SNPRINTF(buffer,sizeof(buffer))  _("execv '%s' failed - %s"), l.list[0],
			Errormsg(errno) );
		Write_fd_str(2,buffer);
		exit(JABORT);
	}

	close(error_fd[1]);
	Free_line_list(&l);
	SNPRINTF(buffer, sizeof(buffer))  "%s", pgm );
	if( (s = safestrpbrk( buffer, Whitespace )) ) *s = 0;
	if( (s = safestrrchr( buffer, '/' )) ){
		++s;
	} else {
		s = buffer;
	}
	LOGMSG(_("Filter_file: started %s- '%s'"), title, s);
	len = 0;
	buffer[0] = 0;
	if( errbuffer ) errbuffer[0] = 0;
	while( len < sizeof(buffer)-1
		&& (n = read(error_fd[0],buffer+len,sizeof(buffer)-1-len)) > 0 ){
		buffer[n+len] = 0;
		while( (s = safestrchr(buffer,'\n')) ){
			*s++ = 0;
			LOGMSG(_("Filter_file: '%s' error msg '%s'"), title, buffer );
			memmove(buffer,s,strlen(s)+1);
		}
		len = strlen(buffer);
	}
	if( len ){
		LOGMSG(_("Filter_file: '%s' error msg '%s'"), title, buffer );
	}
	close(error_fd[0]);
	while( (n = waitpid(pid,&status,0)) != pid ){
		if( pid == -1 && errno == ECHILD ){
			Errorcode = JABORT;
			LOGERR_DIE( _("ifhp: waitpid and no child!") );
		}
		DEBUG1("Filter_file: converter pid %d, exit '%s'", pid,
			Decode_status( &status ) );
	}
	DEBUG1("Filter_file: converter pid %d, exit '%s'", pid,
		Decode_status( &status ) );
	if( WIFEXITED(status) ){
		n = WEXITSTATUS(status);
		if( n > 0 && n < 32 ) n += JFAIL-1;
	} else if( WIFSIGNALED(status) ){
		n = WTERMSIG(status);
		FATAL(_("Filter_file: converter process died with signal %d, '%s'"),
			n, Sigstr(n));
		n = JSIGNAL;
	}
	if( n && errbuffer && !errbuffer[0] ){
		SNPRINTF(errbuffer,len_errbuffer) "Converter exited with status '%s'",
		Decode_status( &status ) );
	}
	DEBUG1("Filter_file: freeing pgm");
	if( pgm ) free(pgm); pgm = 0;
	DEBUG1("Filter_file: done freeing pgm");
	return( n );
}

/*
 * Write_error_msg( int fd, char *id )
 *  - we expand and find the values for writing error messages
 *  This error message will actually then be sent to the printer
 *  There are two formats for message:
 *   ps_${id} = ... \%{msg}
 *   pcl_${id} = ... \%{msg}
 *   ${id} = ... \%{msg}
 *  If PostScript available and ps_${id} then we use ps_${id}
 *  else if PCL available and pcl_${id} then we use pcl_${id}
 *  else we use ${id}
 */

void Write_error_msg( int fd, char *id )
{
	char *s;
	char name[SMALLBUFFER];
	/* now we find the string to use */
	s = 0;
	if( !s && Ps ){
		SNPRINTF(name,sizeof(name)) "ps_%s", id );
		 s = Find_str_value( &Model, name, Value_sep );
	}
	if( !s && Pcl ){
		SNPRINTF(name,sizeof(name)) "pcl_%s", id );
		 s = Find_str_value( &Model, name, Value_sep );
	}
	if( !s ){
		SNPRINTF(name,sizeof(name)) "%s", id );
		 s = Find_str_value( &Model, name, Value_sep );
	}
	if( s ){
		/* now we expand it */
		s = Fix_option_str( s, 0, 0, 0 );
		DEBUG3("Write_error_msg: '%s'", s );
		Write_fd_str( fd , s );
	}
}
